from __future__ import annotations

"""Perk ids and runtime metadata extracted from `perks_init_database` (FUN_0042fd90)."""

from dataclasses import dataclass
from enum import IntEnum, IntFlag


class PerkFlags(IntFlag):
    # `perk_can_offer` uses these bits as mode-specific allow-lists:
    # - in quest mode, only perks with QUEST_MODE_ALLOWED are eligible
    # - in two-player mode, only perks with TWO_PLAYER_ALLOWED are eligible
    QUEST_MODE_ALLOWED = 0x1
    TWO_PLAYER_ALLOWED = 0x2
    STACKABLE = 0x4  # can be offered even if already owned


# Native perk metadata defaults to both quest and two-player allow bits enabled
# (`sub_42fac0` initializes flags to 3).
PERK_DEFAULT_FLAGS = PerkFlags.QUEST_MODE_ALLOWED | PerkFlags.TWO_PLAYER_ALLOWED


class PerkId(IntEnum):
    ANTIPERK = 0
    BLOODY_MESS_QUICK_LEARNER = 1
    SHARPSHOOTER = 2
    FASTLOADER = 3
    LEAN_MEAN_EXP_MACHINE = 4
    LONG_DISTANCE_RUNNER = 5
    PYROKINETIC = 6
    INSTANT_WINNER = 7
    GRIM_DEAL = 8
    ALTERNATE_WEAPON = 9
    PLAGUEBEARER = 10
    EVIL_EYES = 11
    AMMO_MANIAC = 12
    RADIOACTIVE = 13
    FASTSHOT = 14
    FATAL_LOTTERY = 15
    RANDOM_WEAPON = 16
    MR_MELEE = 17
    ANXIOUS_LOADER = 18
    FINAL_REVENGE = 19
    TELEKINETIC = 20
    PERK_EXPERT = 21
    UNSTOPPABLE = 22
    REGRESSION_BULLETS = 23
    INFERNAL_CONTRACT = 24
    POISON_BULLETS = 25
    DODGER = 26
    BONUS_MAGNET = 27
    URANIUM_FILLED_BULLETS = 28
    DOCTOR = 29
    MONSTER_VISION = 30
    HOT_TEMPERED = 31
    BONUS_ECONOMIST = 32
    THICK_SKINNED = 33
    BARREL_GREASER = 34
    AMMUNITION_WITHIN = 35
    VEINS_OF_POISON = 36
    TOXIC_AVENGER = 37
    REGENERATION = 38
    PYROMANIAC = 39
    NINJA = 40
    HIGHLANDER = 41
    JINXED = 42
    PERK_MASTER = 43
    REFLEX_BOOSTED = 44
    GREATER_REGENERATION = 45
    BREATHING_ROOM = 46
    DEATH_CLOCK = 47
    MY_FAVOURITE_WEAPON = 48
    BANDAGE = 49
    ANGRY_RELOADER = 50
    ION_GUN_MASTER = 51
    STATIONARY_RELOADER = 52
    MAN_BOMB = 53
    FIRE_CAUGH = 54
    LIVING_FORTRESS = 55
    TOUGH_RELOADER = 56
    LIFELINE_50_50 = 57


@dataclass(frozen=True, slots=True)
class PerkMeta:
    perk_id: PerkId
    name: str
    description: str
    flags: PerkFlags
    prereq: tuple[PerkId, ...] = ()


_PERK_TABLE = [
    PerkMeta(
        perk_id=PerkId.ANTIPERK,
        name="AntiPerk",
        description="You shouldn't be seeing this..",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.BLOODY_MESS_QUICK_LEARNER,
        name="Bloody Mess",
        description="More the merrier. More blood guarantees a 30% better experience. You spill more blood and gain more experience points.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.SHARPSHOOTER,
        name="Sharpshooter",
        description="Miraculously your aiming improves drastically, but you take a little bit more time on actually firing the gun. If you order now, you also get a fancy LASER SIGHT without ANY charge!",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.FASTLOADER,
        name="Fastloader",
        description="Man, you sure know how to load a gun.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.LEAN_MEAN_EXP_MACHINE,
        name="Lean Mean Exp Machine",
        description="Why kill for experience when you can make some of your own for free! With this perk the experience just keeps flowing in at a constant rate.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.LONG_DISTANCE_RUNNER,
        name="Long Distance Runner",
        description="You move like a train that has feet and runs. You just need a little time to warm up. In other words you'll move faster the longer you run without stopping.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.PYROKINETIC,
        name="Pyrokinetic",
        description="You see flames everywhere. Bare aiming at creatures causes them to heat up.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.INSTANT_WINNER,
        name="Instant Winner",
        description="2500 experience points. Right away. Take it or leave it.",
        flags=PERK_DEFAULT_FLAGS | PerkFlags.STACKABLE,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.GRIM_DEAL,
        name="Grim Deal",
        description="I'll make you a deal: I'll give you 18% more experience points, and you'll give me your life. So you'll die but score higher. Ponder that one for a sec.",
        flags=PerkFlags(0),
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.ALTERNATE_WEAPON,
        name="Alternate Weapon",
        description="Ever fancied about having two weapons available for use? This might be your lucky day; with this perk you'll get an extra weapon slot for another gun! Carrying around two guns slows you down slightly though. (You can switch the weapon slots with RELOAD key)",
        flags=PerkFlags.QUEST_MODE_ALLOWED,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.PLAGUEBEARER,
        name="Plaguebearer",
        description="You carry a horrible disease. Good for you: you are immune. Bad for them: it is contagious! (Monsters become resistant over time though.)",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.EVIL_EYES,
        name="Evil Eyes",
        description="No living (nor dead) can resist the hypnotic power of your eyes: monsters freeze still as you look at them!",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.AMMO_MANIAC,
        name="Ammo Maniac",
        description="You squeeze and you push and you pack your clips with about 20% more ammo than a regular fellow. They call you Ammo Maniac with a deep respect in their voices.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.RADIOACTIVE,
        name="Radioactive",
        description="You are the Radioactive-man; you have that healthy green glow around you! Others don't like it though, it makes them sick and nauseous whenever near you. It does affect your social life a bit.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.FASTSHOT,
        name="Fastshot",
        description="Funny how you make your gun spit bullets faster than the next guy. Even the most professional of engineers are astonished.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.FATAL_LOTTERY,
        name="Fatal Lottery",
        description="Fifty-fifty chance of dying OR gaining 10k experience points. Place your bets. Interested, anyone?",
        flags=PerkFlags.STACKABLE,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.RANDOM_WEAPON,
        name="Random Weapon",
        description="Here, have this weapon. No questions asked.",
        flags=PerkFlags.QUEST_MODE_ALLOWED | PerkFlags.STACKABLE,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.MR_MELEE,
        name="Mr. Melee",
        description="You master the art of melee fighting. You don't just stand still when monsters come near -- you hit back. Hard.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.ANXIOUS_LOADER,
        name="Anxious Loader",
        description="When you can't stand waiting your gun to be reloaded you can speed up the process by clicking your FIRE button repeatedly as fast as you can.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.FINAL_REVENGE,
        name="Final Revenge",
        description="Pick this and you'll get your revenge. It's a promise.",
        flags=PerkFlags(0),
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.TELEKINETIC,
        name="Telekinetic",
        description="Picking up bonuses has never been so easy and FUN. You can pick up bonuses simply by aiming at them for a while. Ingenious.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.PERK_EXPERT,
        name="Perk Expert",
        description="You sure know how to pick a perk -- most people just don't see that extra perk laying around. This gives you the opportunity to pick the freshest and shiniest perks from the top.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.UNSTOPPABLE,
        name="Unstoppable",
        description="Monsters can't slow you down with their nasty scratches and bites. It still hurts but you simply ignore the pain.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.REGRESSION_BULLETS,
        name="Regression Bullets",
        description="Attempt to shoot with an empty clip leads to a severe loss of experience. But hey, whatever makes them go down, right?",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.INFERNAL_CONTRACT,
        name="Infernal Contract",
        description="In exchange for your soul, a dark stranger is offering you three (3) new perks. To collect his part of the bargain soon enough, your health is reduced to a near-death status. Just sign down here below this pentagram..",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.POISON_BULLETS,
        name="Poison Bullets",
        description="You tend to explicitly treat each of your bullets with rat poison. You do it for good luck, but it seems to have other side effects too.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.DODGER,
        name="Dodger",
        description="It seems so stupid just to take the hits. Each time a monster attacks you you have a chance to dodge the attack.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.BONUS_MAGNET,
        name="Bonus Magnet",
        description="You somehow seem to lure all kinds of bonuses to appear around you more often.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.URANIUM_FILLED_BULLETS,
        name="Uranium Filled Bullets",
        description="Your bullets have a nice creamy uranium filling. Yummy. Now that's gotta hurt the monsters more, right?",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.DOCTOR,
        name="Doctor",
        description="With a single glance you can tell the medical condition of, well, anything. Also, being a doctor, you know exactly what hurts the most enabling you to do slightly more damage with your attacks.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.MONSTER_VISION,
        name="Monster Vision",
        description="With your newly enhanced senses you can see all bad energy VERY clearly. That's got to be enough.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.HOT_TEMPERED,
        name="Hot Tempered",
        description="It literally boils inside you. That's exactly why you need to let it out once in a while, unfortunately for those near you.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.BONUS_ECONOMIST,
        name="Bonus Economist",
        description="Your bonus power-ups last 50% longer than they normally would.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.THICK_SKINNED,
        name="Thick Skinned",
        description="Trade 1/3 of your health for only receiving 2/3rds damage on attacks.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.BARREL_GREASER,
        name="Barrel Greaser",
        description="After studying a lot of physics and friction you've come up with a way to make your bullets fly faster. More speed, more damage.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.AMMUNITION_WITHIN,
        name="Ammunition Within",
        description="Empty clip doesn't prevent you from shooting with a weapon; instead the ammunition is drawn from your health while you are reloading.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.VEINS_OF_POISON,
        name="Veins of Poison",
        description="A strong poison runs through your veins. Monsters taking a bite of you are eventually to experience an agonizing death.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.TOXIC_AVENGER,
        name="Toxic Avenger",
        description="You started out just by being poisonous. The next logical step for you is to become highly toxic -- the ULTIMATE TOXIC AVENGER. Most monsters touching you will just drop dead within seconds!",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(PerkId.VEINS_OF_POISON,),
    ),
    PerkMeta(
        perk_id=PerkId.REGENERATION,
        name="Regeneration",
        description="Your health replenishes but very slowly. What more there is to say?",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.PYROMANIAC,
        name="Pyromaniac",
        description="You just enjoy using fire as your Tool of Destruction and you're good at it too; your fire based weapons do a lot more damage.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.NINJA,
        name="Ninja",
        description="You've taken your dodging abilities to the next level; monsters have really hard time hitting you.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(PerkId.DODGER,),
    ),
    PerkMeta(
        perk_id=PerkId.HIGHLANDER,
        name="Highlander",
        description="You are immortal. Well, almost immortal. Instead of actually losing health on attacks you've got a 10% chance of just dropping dead whenever a monster attacks you. There really can be only one, you know.",
        flags=PerkFlags(0),
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.JINXED,
        name="Jinxed",
        description="Things happen near you. Strangest things. Creatures just drop dead and accidents happen. Beware.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.PERK_MASTER,
        name="Perk Master",
        description="Being the Perk Expert taught you a few things and now you are ready to take your training to the next level doubling the ability effect.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(PerkId.PERK_EXPERT,),
    ),
    PerkMeta(
        perk_id=PerkId.REFLEX_BOOSTED,
        name="Reflex Boosted",
        description="To you the world seems to go on about 10% slower than to an average person. It can be rather irritating sometimes, but it does give you a chance to react better.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.GREATER_REGENERATION,
        name="Greater Regeneration",
        description="Your health replenishes faster than ever.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(PerkId.REGENERATION,),
    ),
    PerkMeta(
        perk_id=PerkId.BREATHING_ROOM,
        name="Breathing Room",
        description="Trade 2/3rds of your health for the killing of every single creature on the screen. No, you don't get the experience.",
        flags=PerkFlags.TWO_PLAYER_ALLOWED,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.DEATH_CLOCK,
        name="Death Clock",
        description="You die exactly in 30 seconds. You can't escape your destiny, but feel free to go on a spree. Tick, tock.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.MY_FAVOURITE_WEAPON,
        name="My Favourite Weapon",
        description="You've grown very fond of your piece. You polish it all the time and talk nice to it, your precious. (+2 clip size, no more random weapon bonuses)",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.BANDAGE,
        name="Bandage",
        description="Here, eat this bandage and you'll feel a lot better in no time. (restores up to 50% health)",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.ANGRY_RELOADER,
        name="Angry Reloader",
        description="You hate it when you run out of shots. You HATE HATE HATE reloading your gun. Lucky for you, and strangely enough, your hate materializes as Mighty Balls of Fire. Or more like Quite Decent Balls of Fire, but it's still kinda neat, huh?",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.ION_GUN_MASTER,
        name="Ion Gun Master",
        description="You're good with ion weapons. You're so good that not only your shots do slightly more damage but your ion blast radius is also increased.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.STATIONARY_RELOADER,
        name="Stationary Reloader",
        description="It's incredibly hard to reload your piece while moving around, you've noticed. In fact, realizing that, when you don't move a (leg) muscle you can reload the gun THREE TIMES FASTER!",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.MAN_BOMB,
        name="Man Bomb",
        description="You have the ability to go boom for you are the MAN BOMB. Going boom requires a lot of concentration and standing completely still for a few seconds.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.FIRE_CAUGH,
        name="Fire Caugh",
        description="You have a fireball stuck in your throat. Repeatedly. Mind your manners.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.LIVING_FORTRESS,
        name="Living Fortress",
        description="It comes a time in each man's life when you'd just rather not move anymore. Being living fortress not moving comes with extra benefits as well. You do the more damage the longer you stand still.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.TOUGH_RELOADER,
        name="Tough Reloader",
        description="Damage received during reloading a weapon is halved.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
    PerkMeta(
        perk_id=PerkId.LIFELINE_50_50,
        name="Lifeline 50-50",
        description="The computer removes half of the wrong monsters for you. You don't gain any experience.",
        flags=PERK_DEFAULT_FLAGS,
        prereq=(),
    ),
]

PERK_BY_ID = {int(entry.perk_id): entry for entry in _PERK_TABLE}

QUICK_LEARNER_NAME = "Quick Learner"
QUICK_LEARNER_DESCRIPTION = (
    "You learn things faster than a regular Joe from now on gaining 30% more experience points from everything you do."
)

_PERK_FIXED_NAMES = {
    int(PerkId.FIRE_CAUGH): "Fire Cough",
}

_PERK_FIXED_DESCRIPTIONS = {
    int(PerkId.ANXIOUS_LOADER): "When you can't stand waiting for your gun to be reloaded you can speed up the process by clicking your FIRE button repeatedly as fast as you can.",
    int(PerkId.PERK_EXPERT): "You sure know how to pick a perk -- most people just don't see that extra perk laying around. This gives you the opportunity to pick the freshest and shiniest perks from the top.",
    int(PerkId.DODGER): "It seems so stupid just to take the hits. Each time a monster attacks you, you have a chance to dodge the attack.",
    int(PerkId.NINJA): "You've taken your dodging abilities to the next level; monsters have a really hard time hitting you.",
    int(PerkId.LIVING_FORTRESS): "There comes a time in each man's life when you'd just rather not move anymore. Being a living fortress comes with extra benefits as well. You do more damage the longer you stand still.",
}


def perk_display_name(perk_id: int, *, fx_toggle: int = 0, preserve_bugs: bool = False) -> str:
    if int(perk_id) == int(PerkId.BLOODY_MESS_QUICK_LEARNER) and int(fx_toggle) != 0:
        return QUICK_LEARNER_NAME
    entry = PERK_BY_ID.get(int(perk_id))
    if entry is None:
        return f"Perk {int(perk_id)}"
    if not bool(preserve_bugs):
        fixed = _PERK_FIXED_NAMES.get(int(perk_id))
        if fixed is not None:
            return fixed
    return entry.name


def perk_display_description(perk_id: int, *, fx_toggle: int = 0, preserve_bugs: bool = False) -> str:
    if int(perk_id) == int(PerkId.BLOODY_MESS_QUICK_LEARNER) and int(fx_toggle) != 0:
        return QUICK_LEARNER_DESCRIPTION
    entry = PERK_BY_ID.get(int(perk_id))
    if entry is None:
        return "Unknown perk."
    if not bool(preserve_bugs):
        fixed = _PERK_FIXED_DESCRIPTIONS.get(int(perk_id))
        if fixed is not None:
            return fixed
    return entry.description


def perk_label(perk_id: int, *, fx_toggle: int = 0, preserve_bugs: bool = False) -> str:
    return perk_display_name(perk_id, fx_toggle=fx_toggle, preserve_bugs=preserve_bugs)



from __future__ import annotations

from dataclasses import dataclass

import pyray as rl

from grim.fonts.small import SmallFontData, load_small_font
from grim.geom import Vec2
from grim.view import View, ViewContext

from ._ui_helpers import draw_ui_text, ui_line_height
from .registry import register_view

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)


@dataclass(frozen=True, slots=True)
class UiTexture:
    name: str
    texture: rl.Texture


class UiTextureView:
    def __init__(self, ctx: ViewContext) -> None:
        self._assets_root = ctx.assets_dir
        self._missing_assets: list[str] = []
        self._textures: list[UiTexture] = []
        self._index = 0
        self._small: SmallFontData | None = None

    def open(self) -> None:
        self._missing_assets.clear()
        self._textures.clear()
        self._small = load_small_font(self._assets_root)
        ui_dir = self._assets_root / "crimson" / "ui"
        if not ui_dir.is_dir():
            self._missing_assets.append("ui/")
            raise FileNotFoundError(f"Missing UI assets directory: {ui_dir}")
        for path in sorted(ui_dir.glob("*.png")):
            texture = rl.load_texture(str(path))
            self._textures.append(UiTexture(name=path.name, texture=texture))

    def close(self) -> None:
        for entry in self._textures:
            rl.unload_texture(entry.texture)
        self._textures.clear()
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None

    def update(self, dt: float) -> None:
        del dt

    def _advance(self, delta: int) -> None:
        if not self._textures:
            return
        self._index = (self._index + delta) % len(self._textures)

    def _handle_input(self) -> None:
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
            self._advance(1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
            self._advance(-1)

    def draw(self) -> None:
        rl.clear_background(rl.Color(12, 12, 14, 255))
        if self._missing_assets:
            message = "Missing assets: " + ", ".join(self._missing_assets)
            draw_ui_text(self._small, message, Vec2(24, 24), scale=UI_TEXT_SCALE, color=UI_ERROR_COLOR)
            return
        if not self._textures:
            draw_ui_text(self._small, "No UI textures loaded.", Vec2(24, 24), scale=UI_TEXT_SCALE, color=UI_TEXT_COLOR)
            return

        self._handle_input()
        entry = self._textures[self._index]

        margin = 24
        header_y = margin
        line_height = ui_line_height(self._small, scale=UI_TEXT_SCALE)
        title = f"{self._index + 1}/{len(self._textures)} {entry.name}"
        draw_ui_text(self._small, title, Vec2(margin, header_y), scale=UI_TEXT_SCALE, color=UI_TEXT_COLOR)
        header_y += line_height + 6
        draw_ui_text(
            self._small, "Left/Right: texture", Vec2(margin, header_y), scale=UI_TEXT_SCALE, color=UI_HINT_COLOR,
        )

        available_width = rl.get_screen_width() - margin * 2
        available_height = rl.get_screen_height() - (header_y + line_height + margin)
        scale = min(
            2.0,
            available_width / entry.texture.width,
            available_height / entry.texture.height,
        )
        draw_w = entry.texture.width * scale
        draw_h = entry.texture.height * scale
        x = margin + (available_width - draw_w) / 2
        y = header_y + line_height + (available_height - draw_h) / 2

        src = rl.Rectangle(0.0, 0.0, float(entry.texture.width), float(entry.texture.height))
        dst = rl.Rectangle(float(x), float(y), float(draw_w), float(draw_h))
        rl.draw_texture_pro(entry.texture, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)


@register_view("ui", "UI texture preview")
def build_ui_view(ctx: ViewContext) -> View:
    return UiTextureView(ctx)



from __future__ import annotations

import math

import pyray as rl
from raylib import defines as rd

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp

from ...projectiles import ProjectileTypeId
from ...sim.world_defs import KNOWN_PROJ_FRAMES
from .common import proj_origin
from .types import ProjectileDrawCtx


def _begin_darken_src_zero_blend() -> None:
    # Native projectile_render switches Plague Spreader to D3D8 SRC=ZERO / DST=INVSRCALPHA.
    rl.rl_set_blend_factors_separate(
        rd.RL_ZERO,
        rd.RL_ONE_MINUS_SRC_ALPHA,
        rd.RL_ZERO,
        rd.RL_ONE,
        rd.RL_FUNC_ADD,
        rd.RL_FUNC_ADD,
    )
    rl.begin_blend_mode(rl.BlendMode.BLEND_CUSTOM_SEPARATE)
    rl.rl_set_blend_factors_separate(
        rd.RL_ZERO,
        rd.RL_ONE_MINUS_SRC_ALPHA,
        rd.RL_ZERO,
        rd.RL_ONE,
        rd.RL_FUNC_ADD,
        rd.RL_FUNC_ADD,
    )


def draw_pulse_gun(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    if int(ctx.type_id) != int(ProjectileTypeId.PULSE_GUN):
        return False
    if ctx.texture is None:
        return False

    mapping = KNOWN_PROJ_FRAMES.get(int(ctx.type_id))
    if mapping is None:
        return True
    grid, frame = mapping
    cell_w = float(ctx.texture.width) / float(grid)

    alpha = float(ctx.alpha)
    life = float(ctx.life)
    if life >= 0.4:
        origin = proj_origin(ctx.proj, ctx.pos)
        dist = origin.distance_to(ctx.pos)

        desired_size = dist * 0.16 * ctx.scale
        if desired_size <= 1e-3:
            return True
        sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
        if sprite_scale <= 1e-6:
            return True

        tint = RGBA(0.1, 0.6, 0.2, alpha * 0.7).to_rl()
        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
        renderer._draw_atlas_sprite(
            ctx.texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=sprite_scale,
            rotation_rad=ctx.angle,
            tint=tint,
        )
        rl.end_blend_mode()
        return True

    fade = clamp(life * 2.5, 0.0, 1.0)
    fade_alpha = fade * alpha
    if fade_alpha <= 1e-3:
        return True

    desired_size = 56.0 * ctx.scale
    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    tint = RGBA(1.0, 1.0, 1.0, fade_alpha).to_rl()
    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
    renderer._draw_atlas_sprite(
        ctx.texture,
        grid=grid,
        frame=frame,
        pos=ctx.screen_pos,
        scale=sprite_scale,
        rotation_rad=ctx.angle,
        tint=tint,
    )
    rl.end_blend_mode()
    return True


def draw_splitter_or_blade(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    type_id = int(ctx.type_id)
    if type_id not in (int(ProjectileTypeId.SPLITTER_GUN), int(ProjectileTypeId.BLADE_GUN)):
        return False
    if ctx.texture is None:
        return False

    mapping = KNOWN_PROJ_FRAMES.get(type_id)
    if mapping is None:
        return True
    grid, frame = mapping
    cell_w = float(ctx.texture.width) / float(grid)

    if float(ctx.life) < 0.4:
        return True

    origin = proj_origin(ctx.proj, ctx.pos)
    dist = origin.distance_to(ctx.pos)

    desired_size = min(dist, 20.0) * ctx.scale
    if desired_size <= 1e-3:
        return True

    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    rotation_rad = ctx.angle
    rgb = (1.0, 1.0, 1.0)
    if type_id == int(ProjectileTypeId.BLADE_GUN):
        rotation_rad = float(int(ctx.proj_index)) * 0.1 - float(renderer.elapsed_ms) * 0.1
        rgb = (0.8, 0.8, 0.8)

    tint = RGBA(rgb[0], rgb[1], rgb[2], float(ctx.alpha)).to_rl()
    renderer._draw_atlas_sprite(
        ctx.texture,
        grid=grid,
        frame=frame,
        pos=ctx.screen_pos,
        scale=sprite_scale,
        rotation_rad=rotation_rad,
        tint=tint,
    )
    return True


def draw_plague_spreader(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    if int(ctx.type_id) != int(ProjectileTypeId.PLAGUE_SPREADER):
        return False
    texture = ctx.texture
    if texture is None:
        return False

    grid = 4
    frame = 2
    cell_w = float(texture.width) / float(grid)

    alpha = float(ctx.alpha)
    life = float(ctx.life)
    if life >= 0.4:
        tint = RGBA(1.0, 1.0, 1.0, alpha).to_rl()

        def draw_plague_quad(*, pos: Vec2, size: float) -> None:
            size = float(size)
            if size <= 1e-3:
                return
            desired_size = size * ctx.scale
            sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
            if sprite_scale <= 1e-6:
                return
            pos_screen = renderer.world_to_screen(pos)
            renderer._draw_atlas_sprite(
                texture,
                grid=grid,
                frame=frame,
                pos=pos_screen,
                scale=sprite_scale,
                rotation_rad=0.0,
                tint=tint,
            )

        _begin_darken_src_zero_blend()
        try:
            draw_plague_quad(pos=ctx.pos, size=60.0)

            offset = Vec2.from_heading(ctx.angle + math.pi) * 15.0
            draw_plague_quad(
                pos=ctx.pos + offset,
                size=60.0,
            )

            phase = float(int(ctx.proj_index)) + float(renderer.elapsed_ms) * 0.01
            cos_phase = math.cos(phase)
            sin_phase = math.sin(phase)
            draw_plague_quad(
                pos=ctx.pos.offset(dx=cos_phase * cos_phase - 5.0, dy=sin_phase * 11.0 - 5.0),
                size=52.0,
            )

            phase_120 = phase + 2.0943952
            sin_phase_120 = math.sin(phase_120)
            draw_plague_quad(
                pos=ctx.pos + Vec2.from_polar(phase_120, 10.0),
                size=62.0,
            )

            phase_240 = phase + 4.1887903
            draw_plague_quad(
                pos=ctx.pos + Vec2(math.cos(phase_240) * 10.0, math.sin(phase_240) * sin_phase_120),
                size=62.0,
            )
        finally:
            rl.end_blend_mode()
        return True

    fade = clamp(life * 2.5, 0.0, 1.0)
    fade_alpha = fade * alpha
    if fade_alpha <= 1e-3:
        return True

    desired_size = (fade * 40.0 + 32.0) * ctx.scale
    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    tint = RGBA(1.0, 1.0, 1.0, fade_alpha).to_rl()
    _begin_darken_src_zero_blend()
    try:
        renderer._draw_atlas_sprite(
            texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=sprite_scale,
            rotation_rad=0.0,
            tint=tint,
        )
    finally:
        rl.end_blend_mode()
    return True


__all__ = ["draw_plague_spreader", "draw_pulse_gun", "draw_splitter_or_blade"]



# type: ignore[invalid-assignment,invalid-argument-type,no-matching-overload]
from __future__ import annotations

import math
from collections.abc import Sequence
from dataclasses import asdict, dataclass

from ..replay.checkpoints import ReplayCheckpoint

DEFAULT_RNG_MARK_ORDER: tuple[str, ...] = (
    "before_world_step",
    "gw_begin",
    "gw_after_weapon_refresh",
    "gw_after_perks_rebuild",
    "gw_after_time_scale",
    "ws_begin",
    "ws_after_perk_effects",
    "ws_after_effects_update",
    "ws_after_creatures",
    "ws_after_projectiles",
    "ws_after_secondary_projectiles",
    "ws_after_particles_update",
    "ws_after_sprite_effects",
    "ws_after_particles",
    "ws_after_player_update_p0",
    "ws_after_player_update",
    "ws_after_bonus_update",
    "ws_after_progression",
    "ws_after_sfx_queue_merge",
    "ws_after_player_damage_sfx",
    "ws_after_sfx",
    "after_world_step",
    "after_stage_spawns",
    "after_wave_spawns",
    "after_rush_spawns",
)


@dataclass(frozen=True, slots=True)
class ReplayDiffFailure:
    kind: str
    tick_index: int
    expected: ReplayCheckpoint
    actual: ReplayCheckpoint | None = None
    first_rng_mark: str | None = None


@dataclass(frozen=True, slots=True)
class ReplayDiffResult:
    ok: bool
    checked_count: int
    first_rng_only_tick: int | None = None
    failure: ReplayDiffFailure | None = None


@dataclass(frozen=True, slots=True)
class ReplayFieldDiff:
    field: str
    expected: object
    actual: object


def _checkpoint_to_obj(
    checkpoint: ReplayCheckpoint,
    *,
    include_hash_fields: bool,
    include_rng_fields: bool,
) -> dict[str, object]:
    obj = asdict(checkpoint)
    if not include_hash_fields:
        for key in ("state_hash", "command_hash"):
            obj.pop(key, None)
    if not include_rng_fields:
        for key in ("rng_state", "rng_marks"):
            obj.pop(key, None)
    return obj


def _int_is_unknown(value: object) -> bool:
    return isinstance(value, int) and int(value) < 0


def normalize_unknown_fields(exp: dict[str, object], act: dict[str, object]) -> None:
    for key in ("elapsed_ms", "score_xp", "kills", "creature_count", "perk_pending"):
        if _int_is_unknown(exp.get(key)):
            exp[key] = act.get(key)

    exp_bonus = exp.get("bonus_timers")
    act_bonus = act.get("bonus_timers")
    if isinstance(exp_bonus, dict) and isinstance(act_bonus, dict):
        for key, value in list(exp_bonus.items()):
            if _int_is_unknown(value):
                exp_bonus[key] = act_bonus.get(key)

    exp_perk = exp.get("perk")
    act_perk = act.get("perk")
    if isinstance(exp_perk, dict) and isinstance(act_perk, dict):
        if _int_is_unknown(exp_perk.get("pending_count")):
            exp["perk"] = act_perk

    exp_deaths = exp.get("deaths")
    if isinstance(exp_deaths, list) and len(exp_deaths) == 1:
        row = exp_deaths[0]
        if isinstance(row, dict):
            is_unknown_death = (
                _int_is_unknown(row.get("creature_index"))
                and _int_is_unknown(row.get("type_id"))
                and _int_is_unknown(row.get("xp_awarded"))
            )
            if is_unknown_death:
                exp["deaths"] = act.get("deaths")


def _path_join(path: str, suffix: str) -> str:
    if not path:
        return suffix
    return f"{path}.{suffix}"


def _values_equal(expected: object, actual: object, *, float_abs_tol: float) -> bool:
    abs_tol = max(0.0, float(float_abs_tol)) + 1e-12
    if isinstance(expected, float) and isinstance(actual, (int, float)):
        return math.isclose(float(expected), float(actual), rel_tol=0.0, abs_tol=abs_tol)
    if isinstance(actual, float) and isinstance(expected, (int, float)):
        return math.isclose(float(expected), float(actual), rel_tol=0.0, abs_tol=abs_tol)
    return expected == actual


def _collect_field_diffs(
    *,
    path: str,
    expected: object,
    actual: object,
    out: list[ReplayFieldDiff],
    max_diffs: int | None,
    float_abs_tol: float,
) -> None:
    if max_diffs is not None and len(out) >= int(max_diffs):
        return

    if isinstance(expected, dict) and isinstance(actual, dict):
        keys = sorted({*expected.keys(), *actual.keys()})
        for key in keys:
            key_str = str(key)
            exp_value = expected.get(key_str, expected.get(key))
            act_value = actual.get(key_str, actual.get(key))
            has_exp = key in expected or key_str in expected
            has_act = key in actual or key_str in actual
            if not has_exp or not has_act:
                out.append(
                    ReplayFieldDiff(
                        field=_path_join(path, key_str),
                        expected=exp_value if has_exp else "<missing>",
                        actual=act_value if has_act else "<missing>",
                    ),
                )
                if max_diffs is not None and len(out) >= int(max_diffs):
                    return
                continue
            _collect_field_diffs(
                path=_path_join(path, key_str),
                expected=exp_value,
                actual=act_value,
                out=out,
                max_diffs=max_diffs,
                float_abs_tol=float_abs_tol,
            )
            if max_diffs is not None and len(out) >= int(max_diffs):
                return
        return

    if isinstance(expected, list) and isinstance(actual, list):
        if len(expected) != len(actual):
            out.append(
                ReplayFieldDiff(
                    field=_path_join(path, "_len"),
                    expected=int(len(expected)),
                    actual=int(len(actual)),
                ),
            )
            if max_diffs is not None and len(out) >= int(max_diffs):
                return
        for idx, (exp_value, act_value) in enumerate(zip(expected, actual)):
            _collect_field_diffs(
                path=f"{path}[{idx}]" if path else f"[{idx}]",
                expected=exp_value,
                actual=act_value,
                out=out,
                max_diffs=max_diffs,
                float_abs_tol=float_abs_tol,
            )
            if max_diffs is not None and len(out) >= int(max_diffs):
                return
        return

    # Capture checkpoints quantize global bonus timers to integer ms in JS.
    # A one-ms jitter can appear from float edge cases and self-heal on the
    # next tick without affecting deterministic simulation behavior.
    if path.startswith("bonus_timers.") and isinstance(expected, int) and isinstance(actual, int):
        timer_key = path.removeprefix("bonus_timers.")
        if timer_key in {"2", "4", "6", "9", "11"}:
            if int(expected) > 0 and int(actual) > 0 and abs(int(expected) - int(actual)) <= 1:
                return

    if not _values_equal(expected, actual, float_abs_tol=float_abs_tol):
        out.append(
            ReplayFieldDiff(
                field=path or "<root>",
                expected=expected,
                actual=actual,
            ),
        )


def checkpoint_field_diffs(
    expected: ReplayCheckpoint,
    actual: ReplayCheckpoint,
    *,
    include_hash_fields: bool = True,
    include_rng_fields: bool = True,
    normalize_unknown: bool = True,
    unknown_events_wildcard: bool = True,
    elapsed_baseline: tuple[int, int] | None = None,
    max_diffs: int | None = None,
    float_abs_tol: float = 0.0001,
) -> list[ReplayFieldDiff]:
    exp_obj = _checkpoint_to_obj(
        expected,
        include_hash_fields=bool(include_hash_fields),
        include_rng_fields=bool(include_rng_fields),
    )
    act_obj = _checkpoint_to_obj(
        actual,
        include_hash_fields=bool(include_hash_fields),
        include_rng_fields=bool(include_rng_fields),
    )

    if elapsed_baseline is not None:
        exp_base, act_base = elapsed_baseline
        exp_elapsed = exp_obj.get("elapsed_ms")
        act_elapsed = act_obj.get("elapsed_ms")
        if isinstance(exp_elapsed, int) and isinstance(act_elapsed, int):
            if int(exp_elapsed) >= 0 and int(act_elapsed) >= 0:
                exp_obj["elapsed_ms"] = int(exp_elapsed) - int(exp_base)
                act_obj["elapsed_ms"] = int(act_elapsed) - int(act_base)

    if normalize_unknown:
        normalize_unknown_fields(exp_obj, act_obj)

    # Legacy sidecars (without `events`) store unknown sentinel values.
    if unknown_events_wildcard and int(expected.events.hit_count) < 0:
        exp_obj["events"] = act_obj.get("events")

    diffs: list[ReplayFieldDiff] = []
    _collect_field_diffs(
        path="",
        expected=exp_obj,
        actual=act_obj,
        out=diffs,
        max_diffs=max_diffs,
        float_abs_tol=float_abs_tol,
    )
    return diffs


def compare_checkpoints(
    expected: Sequence[ReplayCheckpoint],
    actual: Sequence[ReplayCheckpoint],
    *,
    rng_mark_order: Sequence[str] = DEFAULT_RNG_MARK_ORDER,
) -> ReplayDiffResult:
    actual_by_tick = {int(ckpt.tick_index): ckpt for ckpt in actual}
    first_rng_only_tick: int | None = None
    checked_count = 0

    for exp in expected:
        checked_count += 1
        tick = int(exp.tick_index)
        act = actual_by_tick.get(tick)
        if act is None:
            return ReplayDiffResult(
                ok=False,
                checked_count=checked_count,
                first_rng_only_tick=first_rng_only_tick,
                failure=ReplayDiffFailure(
                    kind="missing_checkpoint",
                    tick_index=tick,
                    expected=exp,
                    actual=None,
                ),
            )

        if str(exp.command_hash) and str(exp.command_hash) != str(act.command_hash):
            return ReplayDiffResult(
                ok=False,
                checked_count=checked_count,
                first_rng_only_tick=first_rng_only_tick,
                failure=ReplayDiffFailure(
                    kind="command_mismatch",
                    tick_index=tick,
                    expected=exp,
                    actual=act,
                ),
            )

        if str(exp.state_hash) == str(act.state_hash):
            continue

        exp_no_rng = _checkpoint_to_obj(exp, include_hash_fields=False, include_rng_fields=False)
        act_no_rng = _checkpoint_to_obj(act, include_hash_fields=False, include_rng_fields=False)
        normalize_unknown_fields(exp_no_rng, act_no_rng)
        # Legacy sidecars (without `events`) store unknown sentinel values.
        if int(exp.events.hit_count) < 0:
            exp_no_rng["events"] = act_no_rng.get("events")

        if exp_no_rng == act_no_rng:
            if first_rng_only_tick is None:
                first_rng_only_tick = tick
            continue

        mark_keys = sorted({*exp.rng_marks.keys(), *act.rng_marks.keys()})
        mark_mismatch = [key for key in mark_keys if int(exp.rng_marks.get(key, -1)) != int(act.rng_marks.get(key, -1))]
        first_mark = next((key for key in rng_mark_order if key in mark_mismatch), mark_mismatch[0] if mark_mismatch else None)
        return ReplayDiffResult(
            ok=False,
            checked_count=checked_count,
            first_rng_only_tick=first_rng_only_tick,
            failure=ReplayDiffFailure(
                kind="state_mismatch",
                tick_index=tick,
                expected=exp,
                actual=act,
                first_rng_mark=first_mark,
            ),
        )

    return ReplayDiffResult(
        ok=True,
        checked_count=checked_count,
        first_rng_only_tick=first_rng_only_tick,
        failure=None,
    )



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from ...sim.state_types import GameplayState, PlayerState

ProjectileSpawnFn = Callable[..., int]
OwnerIdForPlayerFn = Callable[[int], int]
OwnerIdForPlayerProjectilesFn = Callable[[GameplayState, int], int]


@dataclass(slots=True)
class PlayerPerkTickCtx:
    state: GameplayState
    player: PlayerState
    players: list[PlayerState] | None
    dt: float
    stationary: bool
    owner_id_for_player: OwnerIdForPlayerFn
    owner_id_for_player_projectiles: OwnerIdForPlayerProjectilesFn
    projectile_spawn: ProjectileSpawnFn



from __future__ import annotations

from dataclasses import dataclass

import pyray as rl

from crimson.creatures.anim import creature_corpse_frame_for_type
from crimson.effects import FxQueue, FxQueueRotated
from crimson.effects_atlas import EffectId
from crimson.render.terrain_fx import FxQueueTextures, bake_fx_queues
from grim.assets import resolve_asset_path
from grim.color import RGBA
from grim.config import ensure_crimson_cfg
from grim.fonts.grim_mono import GrimMonoFont, load_grim_mono_font
from grim.fonts.small import SmallFontData, load_small_font
from grim.geom import Vec2
from grim.terrain_render import GroundRenderer
from grim.view import View, ViewContext

from ..paths import default_runtime_dir
from ..quests import all_quests
from ..quests.types import QuestDefinition
from ..terrain_assets import TerrainTextureId
from ._ui_helpers import draw_ui_text
from .quest_title_overlay import draw_quest_title_overlay
from .registry import register_view

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)


@dataclass(slots=True)
class GroundAssets:
    textures: dict[int, rl.Texture]


TERRAIN_TEXTURES: list[tuple[int, str]] = [
    (TerrainTextureId.Q1_BASE, "ter/ter_q1_base.png"),
    (TerrainTextureId.Q1_OVERLAY, "ter/ter_q1_tex1.png"),
    (TerrainTextureId.Q2_BASE, "ter/ter_q2_base.png"),
    (TerrainTextureId.Q2_OVERLAY, "ter/ter_q2_tex1.png"),
    (TerrainTextureId.Q3_BASE, "ter/ter_q3_base.png"),
    (TerrainTextureId.Q3_OVERLAY, "ter/ter_q3_tex1.png"),
    (TerrainTextureId.Q4_BASE, "ter/ter_q4_base.png"),
    (TerrainTextureId.Q4_OVERLAY, "ter/ter_q4_tex1.png"),
]


class GroundView:
    def __init__(self, ctx: ViewContext) -> None:
        self._assets_root = ctx.assets_dir
        self._missing_assets: list[str] = []
        self._small: SmallFontData | None = None
        self._grim_mono: GrimMonoFont | None = None
        self._assets: GroundAssets | None = None
        self._renderer: GroundRenderer | None = None
        self._camera = Vec2()
        self._quests: list[QuestDefinition] = []
        self._quest_index = 0
        self._terrain_seed: int | None = None
        self._fx_queue = FxQueue()
        self._fx_queue_rotated = FxQueueRotated()
        self._fx_textures: FxQueueTextures | None = None

    def open(self) -> None:
        self._missing_assets.clear()
        self._small = load_small_font(self._assets_root)
        self._grim_mono = load_grim_mono_font(self._assets_root)
        textures: dict[int, rl.Texture] = {}
        for terrain_id, rel_path in TERRAIN_TEXTURES:
            path = resolve_asset_path(self._assets_root, rel_path)
            if path is None:
                self._missing_assets.append(rel_path)
                continue
            textures[terrain_id] = rl.load_texture(str(path))
        if self._missing_assets:
            raise FileNotFoundError(f"Missing ground assets: {', '.join(self._missing_assets)}")
        self._assets = GroundAssets(textures=textures)
        self._quests = all_quests()
        self._fx_queue.clear()
        self._fx_queue_rotated.clear()
        particles_path = resolve_asset_path(self._assets_root, "game/particles.png")
        if particles_path is None:
            self._missing_assets.append("game/particles.png")
        bodyset_path = resolve_asset_path(self._assets_root, "game/bodyset.png")
        if bodyset_path is None:
            self._missing_assets.append("game/bodyset.png")
        if self._missing_assets:
            raise FileNotFoundError(f"Missing ground assets: {', '.join(self._missing_assets)}")
        self._fx_textures = FxQueueTextures(
            particles=rl.load_texture(str(particles_path)),
            bodyset=rl.load_texture(str(bodyset_path)),
        )
        texture_scale, screen_w, screen_h = self._load_runtime_config()
        if self._renderer is not None:
            self._renderer.texture_scale = texture_scale
            self._renderer.screen_width = screen_w
            self._renderer.screen_height = screen_h
        self._quest_index = 0
        self._apply_quest()

    def close(self) -> None:
        if self._assets is not None:
            for texture in self._assets.textures.values():
                rl.unload_texture(texture)
            self._assets = None
        if self._renderer is not None and self._renderer.render_target is not None:
            rl.unload_render_texture(self._renderer.render_target)
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None
        if self._grim_mono is not None:
            rl.unload_texture(self._grim_mono.texture)
            self._grim_mono = None
        if self._fx_textures is not None:
            rl.unload_texture(self._fx_textures.particles)
            rl.unload_texture(self._fx_textures.bodyset)
            self._fx_textures = None
        self._fx_queue.clear()
        self._fx_queue_rotated.clear()

    def update(self, dt: float) -> None:
        speed = 240.0
        movement = Vec2(
            float(rl.is_key_down(rl.KeyboardKey.KEY_A)) - float(rl.is_key_down(rl.KeyboardKey.KEY_D)),
            float(rl.is_key_down(rl.KeyboardKey.KEY_W)) - float(rl.is_key_down(rl.KeyboardKey.KEY_S)),
        )
        self._camera = self._camera + movement * (speed * dt)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
            self._quest_index = (self._quest_index - 1) % max(1, len(self._quests))
            self._apply_quest()
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
            self._quest_index = (self._quest_index + 1) % max(1, len(self._quests))
            self._apply_quest()
        if self._renderer is not None:
            self._renderer.process_pending()
            if rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT):
                mouse = rl.get_mouse_position()
                world_pos = Vec2.from_xy(mouse) - self._camera
                self._fx_queue.add(
                    effect_id=int(EffectId.BLOOD_SPLATTER),
                    pos=world_pos,
                    width=30.0,
                    height=30.0,
                    rotation=0.0,
                    rgba=RGBA(1.0, 1.0, 1.0, 1.0),
                )
            if self._fx_textures is not None and (self._fx_queue.count or self._fx_queue_rotated.count):
                bake_fx_queues(
                    self._renderer,
                    fx_queue=self._fx_queue,
                    fx_queue_rotated=self._fx_queue_rotated,
                    textures=self._fx_textures,
                    corpse_frame_for_type=creature_corpse_frame_for_type,
                )

    def draw(self) -> None:
        rl.clear_background(rl.Color(12, 12, 14, 255))
        if self._missing_assets:
            message = "Missing assets: " + ", ".join(self._missing_assets)
            draw_ui_text(self._small, message, Vec2(24, 24), scale=UI_TEXT_SCALE, color=UI_ERROR_COLOR)
            return
        if self._renderer is None:
            draw_ui_text(
                self._small, "Ground renderer not initialized.", Vec2(24, 24), scale=UI_TEXT_SCALE, color=UI_ERROR_COLOR,
            )
            return
        self._renderer.draw(self._camera)
        self._draw_quest_title_overlay()

    def _load_runtime_config(self) -> tuple[float, float | None, float | None]:
        runtime_dir = default_runtime_dir()
        if runtime_dir.is_dir():
            try:
                cfg = ensure_crimson_cfg(runtime_dir)
                return (
                    cfg.texture_scale,
                    float(cfg.screen_width),
                    float(cfg.screen_height),
                )
            except (OSError, ValueError):
                return 1.0, None, None
        return 1.0, None, None

    def _apply_quest(self) -> None:
        if not self._quests or self._assets is None:
            return
        quest = self._quests[self._quest_index]
        base_id, overlay_id, detail_id = quest.terrain_ids or (
            TerrainTextureId.Q1_BASE,
            TerrainTextureId.Q1_OVERLAY,
            TerrainTextureId.Q1_BASE,
        )
        textures = self._assets.textures
        base = textures.get(base_id)
        if base is None:
            return
        overlay = textures.get(overlay_id)
        detail = textures.get(detail_id)
        if self._renderer is None:
            texture_scale, screen_w, screen_h = self._load_runtime_config()
            self._renderer = GroundRenderer(
                texture=base,
                overlay=overlay,
                overlay_detail=detail,
                width=1024,
                height=1024,
                texture_scale=texture_scale,
                screen_width=screen_w,
                screen_height=screen_h,
            )
        else:
            self._renderer.texture = base
            self._renderer.overlay = overlay
            self._renderer.overlay_detail = detail
        self._terrain_seed = self._quest_seed(quest.major, quest.minor)
        self._regenerate_terrain(reset_camera=True)

    def _regenerate_terrain(self, *, reset_camera: bool = False) -> None:
        renderer = self._renderer
        if renderer is None:
            return
        renderer.schedule_generate(seed=self._terrain_seed, layers=3)
        if reset_camera:
            self._camera = Vec2()

    def _quest_seed(self, major: int, minor: int) -> int:
        return int(major) * 100 + int(minor)

    def _draw_quest_title_overlay(self) -> None:
        if self._grim_mono is None or not self._quests:
            return
        quest = self._quests[self._quest_index]
        draw_quest_title_overlay(self._grim_mono, quest.title, quest.level)


@register_view("ground", "Ground texture")
def build_ground_view(ctx: ViewContext) -> View:
    return GroundView(ctx)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from ..creatures.spawn import advance_survival_spawn_stage, tick_rush_mode_spawns, tick_survival_wave_spawns
from ..effects import FxQueue, FxQueueRotated
from ..game_modes import GameMode
from ..gameplay import survival_update_weapon_handouts
from ..quests.runtime import tick_quest_completion_transition
from ..quests.timeline import quest_spawn_table_empty, tick_quest_mode_spawns
from ..quests.types import SpawnEntry
from .input import PlayerInput
from .step_pipeline import DeterministicStepResult, run_deterministic_step, time_scale_reflex_boost_bonus
from .world_state import WorldState


@dataclass(slots=True)
class DeterministicSessionTick:
    step: DeterministicStepResult
    elapsed_ms: float
    rng_marks: dict[str, int]
    creature_count_world_step: int


@dataclass(slots=True)
class SurvivalDeterministicSession:
    world: WorldState
    world_size: float
    damage_scale_by_type: dict[int, float]
    fx_queue: FxQueue
    fx_queue_rotated: FxQueueRotated
    detail_preset: int = 5
    fx_toggle: int = 0
    game_tune_started: bool = False
    auto_pick_perks: bool = False
    demo_mode_active: bool = False
    perk_progression_enabled: bool = True
    apply_world_dt_steps: bool = True
    clear_fx_queues_each_tick: bool = False
    elapsed_ms: float = 0.0
    stage: int = 0
    spawn_cooldown_ms: float = 0.0

    def step_tick(
        self,
        *,
        dt_frame: float,
        dt_frame_ms_i32: int | None = None,
        inputs: list[PlayerInput] | None,
        trace_rng: bool = False,
    ) -> DeterministicSessionTick:
        dt_frame = float(dt_frame)
        state = self.world.state
        dt_sim = time_scale_reflex_boost_bonus(
            reflex_boost_timer=float(state.bonuses.reflex_boost),
            time_scale_active=bool(state.time_scale_active),
            dt=float(dt_frame),
        )
        dt_sim_ms = float(dt_sim) * 1000.0
        if dt_frame_ms_i32 is not None and int(dt_frame_ms_i32) > 0:
            # Use captured integer ms for native cadence counters when available,
            # then apply reflex scaling with integer semantics.
            base_dt_ms_i32 = int(dt_frame_ms_i32)
            if bool(state.time_scale_active) and float(dt_frame) > 0.0:
                dt_sim_ms = float(max(0, int(float(dt_sim) * 1000.0)))
            else:
                dt_sim_ms = float(base_dt_ms_i32)
        elapsed_before_ms = float(self.elapsed_ms)

        rng_marks: dict[str, int] = {"before_world_step": int(state.rng.state)}
        dt_sim_ms_value = float(dt_sim_ms)

        def _mid_step_spawns() -> None:
            # Native `survival_update` runs after gameplay world updates:
            # - it observes current player XP/level (post-kill award),
            # - it computes spawn interval from the pre-increment survival elapsed timer.
            survival_update_weapon_handouts(
                state,
                self.world.players,
                survival_elapsed_ms=float(elapsed_before_ms),
            )

            player_level = self.world.players[0].level if self.world.players else 1
            stage, milestone_calls = advance_survival_spawn_stage(self.stage, player_level=int(player_level))
            self.stage = stage
            for call in milestone_calls:
                self.world.creatures.spawn_template(
                    int(call.template_id),
                    call.pos,
                    float(call.heading),
                    state.rng,
                    rand=state.rng.rand,
                )
            rng_marks["after_stage_spawns"] = int(state.rng.state)

            player_xp = self.world.players[0].experience if self.world.players else 0
            cooldown, wave_spawns = tick_survival_wave_spawns(
                self.spawn_cooldown_ms,
                dt_sim_ms_value,
                state.rng,
                player_count=len(self.world.players),
                survival_elapsed_ms=float(elapsed_before_ms),
                player_experience=int(player_xp),
                terrain_width=int(self.world_size),
                terrain_height=int(self.world_size),
            )
            self.spawn_cooldown_ms = cooldown
            self.world.creatures.spawn_inits(wave_spawns)
            rng_marks["after_wave_spawns"] = int(state.rng.state)

        step = run_deterministic_step(
            world=self.world,
            dt_frame=float(dt_frame),
            dt_frame_ms_i32=(int(dt_frame_ms_i32) if dt_frame_ms_i32 is not None else None),
            apply_world_dt_steps=bool(self.apply_world_dt_steps),
            inputs=inputs,
            world_size=float(self.world_size),
            damage_scale_by_type=self.damage_scale_by_type,
            detail_preset=int(self.detail_preset),
            fx_toggle=int(self.fx_toggle),
            fx_queue=self.fx_queue,
            fx_queue_rotated=self.fx_queue_rotated,
            auto_pick_perks=bool(self.auto_pick_perks),
            game_mode=int(GameMode.SURVIVAL),
            demo_mode_active=bool(self.demo_mode_active),
            perk_progression_enabled=bool(self.perk_progression_enabled),
            game_tune_started=bool(self.game_tune_started),
            defer_camera_shake_update=False,
            mid_step_hook=_mid_step_spawns,
            rng_marks_out=rng_marks,
            trace_presentation_rng=bool(trace_rng),
        )
        if step.presentation.trigger_game_tune:
            self.game_tune_started = True

        if self.clear_fx_queues_each_tick:
            # Live gameplay clears terrain FX queues during render (`bake_fx_queues(clear=True)`).
            # Headless verification has no render pass, so clear explicitly per simulated tick.
            self.fx_queue.clear()
            self.fx_queue_rotated.clear()

        creature_count_world_step = sum(1 for creature in self.world.creatures.entries if creature.active)
        rng_marks["after_world_step"] = int(state.rng.state)
        rng_marks["after_camera_update"] = int(rng_marks.get("ws_after_camera_update", state.rng.state))
        self.world.creatures.finalize_post_render_lifecycle()
        self.elapsed_ms = float(elapsed_before_ms) + float(dt_sim_ms)

        return DeterministicSessionTick(
            step=step,
            elapsed_ms=float(self.elapsed_ms),
            rng_marks=rng_marks,
            creature_count_world_step=int(creature_count_world_step),
        )


@dataclass(slots=True)
class RushDeterministicSession:
    world: WorldState
    world_size: float
    damage_scale_by_type: dict[int, float]
    fx_queue: FxQueue
    fx_queue_rotated: FxQueueRotated
    detail_preset: int = 5
    fx_toggle: int = 0
    game_tune_started: bool = False
    clear_fx_queues_each_tick: bool = False
    enforce_loadout: Callable[[], None] | None = None
    elapsed_ms: float = 0.0
    spawn_cooldown_ms: float = 0.0

    def step_tick(
        self,
        *,
        dt_frame: float,
        inputs: list[PlayerInput] | None,
        trace_rng: bool = False,
    ) -> DeterministicSessionTick:
        dt_frame = float(dt_frame)
        dt_frame_ms = float(dt_frame) * 1000.0
        self.elapsed_ms += float(dt_frame_ms)

        if self.enforce_loadout is not None:
            self.enforce_loadout()

        state = self.world.state
        rng_marks: dict[str, int] = {"before_world_step": int(state.rng.state)}

        def _mid_step_spawns() -> None:
            cooldown, spawns = tick_rush_mode_spawns(
                self.spawn_cooldown_ms,
                dt_frame_ms,
                state.rng,
                player_count=len(self.world.players),
                survival_elapsed_ms=int(self.elapsed_ms),
                terrain_width=float(self.world_size),
                terrain_height=float(self.world_size),
            )
            self.spawn_cooldown_ms = cooldown
            self.world.creatures.spawn_inits(spawns)
            rng_marks["after_rush_spawns"] = int(state.rng.state)

        step = run_deterministic_step(
            world=self.world,
            dt_frame=float(dt_frame),
            inputs=inputs,
            world_size=float(self.world_size),
            damage_scale_by_type=self.damage_scale_by_type,
            detail_preset=int(self.detail_preset),
            fx_toggle=int(self.fx_toggle),
            fx_queue=self.fx_queue,
            fx_queue_rotated=self.fx_queue_rotated,
            auto_pick_perks=False,
            game_mode=int(GameMode.RUSH),
            demo_mode_active=False,
            perk_progression_enabled=False,
            game_tune_started=bool(self.game_tune_started),
            defer_camera_shake_update=False,
            mid_step_hook=_mid_step_spawns,
            rng_marks_out=rng_marks,
            trace_presentation_rng=bool(trace_rng),
        )
        if step.presentation.trigger_game_tune:
            self.game_tune_started = True

        if self.clear_fx_queues_each_tick:
            # Live gameplay clears terrain FX queues during render (`bake_fx_queues(clear=True)`).
            # Headless verification has no render pass, so clear explicitly per simulated tick.
            self.fx_queue.clear()
            self.fx_queue_rotated.clear()

        creature_count_world_step = sum(1 for creature in self.world.creatures.entries if creature.active)
        rng_marks["after_world_step"] = int(state.rng.state)
        rng_marks["after_camera_update"] = int(rng_marks.get("ws_after_camera_update", state.rng.state))
        self.world.creatures.finalize_post_render_lifecycle()

        return DeterministicSessionTick(
            step=step,
            elapsed_ms=float(self.elapsed_ms),
            rng_marks=rng_marks,
            creature_count_world_step=int(creature_count_world_step),
        )


@dataclass(slots=True)
class QuestDeterministicSessionTick:
    step: DeterministicStepResult
    elapsed_ms: float
    rng_marks: dict[str, int]
    creature_count_world_step: int
    spawn_timeline_ms: float
    no_creatures_timer_ms: float
    completion_transition_ms: float
    completed: bool
    play_hit_sfx: bool
    play_completion_music: bool


@dataclass(slots=True)
class QuestDeterministicSession:
    world: WorldState
    world_size: float
    damage_scale_by_type: dict[int, float]
    fx_queue: FxQueue
    fx_queue_rotated: FxQueueRotated
    spawn_entries: tuple[SpawnEntry, ...] = ()
    detail_preset: int = 5
    fx_toggle: int = 0
    clear_fx_queues_each_tick: bool = False
    elapsed_ms: float = 0.0
    spawn_timeline_ms: float = 0.0
    no_creatures_timer_ms: float = 0.0
    completion_transition_ms: float = -1.0

    def step_tick(
        self,
        *,
        dt_frame: float,
        inputs: list[PlayerInput] | None,
        trace_rng: bool = False,
    ) -> QuestDeterministicSessionTick:
        dt_frame = float(dt_frame)
        dt_frame_ms = float(dt_frame) * 1000.0
        self.elapsed_ms += float(dt_frame_ms)

        state = self.world.state
        rng_marks: dict[str, int] = {"before_world_step": int(state.rng.state)}

        step = run_deterministic_step(
            world=self.world,
            dt_frame=float(dt_frame),
            inputs=inputs,
            world_size=float(self.world_size),
            damage_scale_by_type=self.damage_scale_by_type,
            detail_preset=int(self.detail_preset),
            fx_toggle=int(self.fx_toggle),
            fx_queue=self.fx_queue,
            fx_queue_rotated=self.fx_queue_rotated,
            auto_pick_perks=False,
            game_mode=int(GameMode.QUESTS),
            demo_mode_active=bool(state.demo_mode_active),
            perk_progression_enabled=True,
            game_tune_started=False,
            defer_camera_shake_update=False,
            rng_marks_out=rng_marks,
            trace_presentation_rng=bool(trace_rng),
        )

        creatures_none_active = not bool(self.world.creatures.iter_active())
        entries, timeline_ms, creatures_none_active, no_creatures_timer_ms, spawns = tick_quest_mode_spawns(
            self.spawn_entries,
            quest_spawn_timeline_ms=float(self.spawn_timeline_ms),
            frame_dt_ms=float(dt_frame_ms),
            terrain_width=float(self.world_size),
            creatures_none_active=bool(creatures_none_active),
            no_creatures_timer_ms=float(self.no_creatures_timer_ms),
        )
        self.spawn_entries = entries
        self.spawn_timeline_ms = float(timeline_ms)
        self.no_creatures_timer_ms = float(no_creatures_timer_ms)
        for call in spawns:
            self.world.creatures.spawn_template(
                int(call.template_id),
                call.pos,
                float(call.heading),
                state.rng,
                rand=state.rng.rand,
            )
        rng_marks["after_quest_spawns"] = int(state.rng.state)

        any_alive_after = any(player.health > 0.0 for player in self.world.players)
        completed = False
        play_hit_sfx = False
        play_completion_music = False
        if any_alive_after:
            completion_ms, completed, play_hit_sfx, play_completion_music = tick_quest_completion_transition(
                float(self.completion_transition_ms),
                frame_dt_ms=float(dt_frame_ms),
                creatures_none_active=bool(creatures_none_active),
                spawn_table_empty=quest_spawn_table_empty(self.spawn_entries),
            )
            self.completion_transition_ms = float(completion_ms)
        else:
            self.completion_transition_ms = -1.0

        if self.clear_fx_queues_each_tick:
            # Live gameplay clears terrain FX queues during render (`bake_fx_queues(clear=True)`).
            # Headless verification has no render pass, so clear explicitly per simulated tick.
            self.fx_queue.clear()
            self.fx_queue_rotated.clear()

        creature_count_world_step = sum(1 for creature in self.world.creatures.entries if creature.active)
        rng_marks["after_world_step"] = int(state.rng.state)
        rng_marks["after_camera_update"] = int(rng_marks.get("ws_after_camera_update", state.rng.state))
        self.world.creatures.finalize_post_render_lifecycle()

        return QuestDeterministicSessionTick(
            step=step,
            elapsed_ms=float(self.elapsed_ms),
            rng_marks=rng_marks,
            creature_count_world_step=int(creature_count_world_step),
            spawn_timeline_ms=float(self.spawn_timeline_ms),
            no_creatures_timer_ms=float(self.no_creatures_timer_ms),
            completion_transition_ms=float(self.completion_transition_ms),
            completed=bool(completed),
            play_hit_sfx=bool(play_hit_sfx),
            play_completion_music=bool(play_completion_music),
        )



from __future__ import annotations

from ..helpers import perk_active
from ..ids import PerkId
from ..runtime.hook_types import PerkHooks
from ..runtime.player_tick_context import PlayerPerkTickCtx


def tick_living_fortress(ctx: PlayerPerkTickCtx) -> None:
    if ctx.stationary and perk_active(ctx.player, PerkId.LIVING_FORTRESS):
        ctx.player.living_fortress_timer = min(30.0, ctx.player.living_fortress_timer + ctx.dt)
    else:
        ctx.player.living_fortress_timer = 0.0


HOOKS = PerkHooks(
    perk_id=PerkId.LIVING_FORTRESS,
    player_tick_steps=(tick_living_fortress,),
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Protocol

import pyray as rl

from grim.geom import Vec2

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ...sim.state_types import PlayerState


class ProjectileLike(Protocol):
    @property
    def origin(self) -> Vec2: ...

    @property
    def speed_scale(self) -> float: ...

    @property
    def base_damage(self) -> float: ...


class SecondaryProjectileLike(Protocol):
    @property
    def detonation_t(self) -> float: ...

    @property
    def detonation_scale(self) -> float: ...


class _CreatureLike(Protocol):
    @property
    def active(self) -> bool: ...

    @property
    def hitbox_size(self) -> float: ...

    @property
    def size(self) -> float: ...

    @property
    def pos(self) -> Vec2: ...


class _CreaturePoolLike(Protocol):
    @property
    def entries(self) -> Sequence[_CreatureLike]: ...


class _FxDetailConfigLike(Protocol):
    def fx_detail(self, *, level: int, default: bool) -> bool: ...


class ProjectileRendererLike(Protocol):
    @property
    def bullet_trail_texture(self) -> rl.Texture | None: ...

    @property
    def bullet_texture(self) -> rl.Texture | None: ...

    @property
    def particles_texture(self) -> rl.Texture | None: ...

    @property
    def projs_texture(self) -> rl.Texture | None: ...

    @property
    def config(self) -> _FxDetailConfigLike | None: ...

    @property
    def players(self) -> Sequence[PlayerState]: ...

    @property
    def creatures(self) -> _CreaturePoolLike: ...

    @property
    def elapsed_ms(self) -> float: ...

    def _is_bullet_trail_type(self, type_id: int) -> bool: ...

    def world_to_screen(self, pos: Vec2) -> Vec2: ...

    def _draw_bullet_trail(
        self,
        start: Vec2,
        end: Vec2,
        *,
        type_id: int,
        alpha: int,
        scale: float,
        angle: float,
    ) -> bool: ...

    def _bullet_sprite_size(self, type_id: int, *, scale: float) -> float: ...

    def _draw_atlas_sprite(
        self,
        texture: rl.Texture,
        *,
        grid: int,
        frame: int,
        pos: Vec2,
        scale: float,
        rotation_rad: float = 0.0,
        tint: rl.Color = rl.WHITE,
    ) -> None: ...


@dataclass(frozen=True, slots=True)
class ProjectileDrawCtx:
    renderer: ProjectileRendererLike
    proj: ProjectileLike
    proj_index: int
    texture: rl.Texture | None
    type_id: int
    pos: Vec2
    screen_pos: Vec2
    life: float
    angle: float
    scale: float
    alpha: float


@dataclass(frozen=True, slots=True)
class SecondaryProjectileDrawCtx:
    renderer: ProjectileRendererLike
    proj: SecondaryProjectileLike
    proj_type: int
    screen_pos: Vec2
    angle: float
    scale: float
    alpha: float


__all__ = [
    "ProjectileDrawCtx",
    "ProjectileLike",
    "ProjectileRendererLike",
    "SecondaryProjectileDrawCtx",
    "SecondaryProjectileLike",
]



from __future__ import annotations

"""
JAZ texture format (Crimsonland).

File layout:
  - u8  method: compression method (1 = zlib)
  - u32 comp_size: compressed payload size (bytes)
  - u32 raw_size: uncompressed payload size (bytes)
  - zlib stream (length = comp_size)

Decompressed payload:
  - u32 jpeg_len
  - jpeg bytes (length = jpeg_len)
  - alpha_rle: (count, value) byte pairs for alpha channel

Notes from assets:
  - alpha runs expand to width*height for most files; one file is short by 1 pixel.
    We pad any remaining pixels with 0 (transparent).
"""

import io
import zlib
from pathlib import Path

from construct import Bytes, Int8ul, Int32ul, Struct, this
from PIL import Image

JAZ_HEADER = Struct(
    "method" / Int8ul,
    "comp_size" / Int32ul,
    "raw_size" / Int32ul,
)

JAZ_FILE = Struct(
    "header" / JAZ_HEADER,
    "compressed" / Bytes(this.header.comp_size),
)


def jaz_payload(raw_size: int) -> Struct:
    return Struct(
        "jpeg_len" / Int32ul,
        "jpeg" / Bytes(this.jpeg_len),
        "alpha_rle" / Bytes(raw_size - 4 - this.jpeg_len),
    )


class JazImage:
    def __init__(self, width: int, height: int, jpeg: bytes, alpha: bytes) -> None:
        self.width = width
        self.height = height
        self.jpeg = jpeg
        self.alpha = alpha

    def rgb_image(self) -> Image.Image:
        img = Image.open(io.BytesIO(self.jpeg))
        return img.convert("RGB")

    def alpha_image(self) -> Image.Image:
        return Image.frombytes("L", (self.width, self.height), self.alpha)

    def composite_image(self) -> Image.Image:
        rgb = self.rgb_image()
        alpha = self.alpha_image()
        rgb.putalpha(alpha)
        return rgb


def decode_alpha_rle(data: bytes, expected: int) -> bytes:
    out = bytearray(expected)
    filled = 0
    for i in range(0, len(data) - 1, 2):
        count = data[i]
        value = data[i + 1]
        if count == 0:
            continue
        if filled >= expected:
            break
        end = min(filled + count, expected)
        out[filled:end] = bytes([value]) * (end - filled)
        filled = end
    return bytes(out)


def decode_jaz_bytes(data: bytes) -> JazImage:
    parsed = JAZ_FILE.parse(data)
    header = parsed.header
    if header.method != 1:
        raise ValueError(f"unsupported compression method: {header.method}")
    raw = zlib.decompress(parsed.compressed)
    if len(raw) != header.raw_size:
        raise ValueError(f"raw size mismatch: {len(raw)} != {header.raw_size}")
    payload = jaz_payload(header.raw_size).parse(raw)
    img = Image.open(io.BytesIO(payload.jpeg))
    width, height = img.size
    alpha = decode_alpha_rle(payload.alpha_rle, width * height)
    return JazImage(width, height, payload.jpeg, alpha)


def decode_jaz(path: str | Path) -> JazImage:
    return decode_jaz_bytes(Path(path).read_bytes())



from __future__ import annotations

import os
from pathlib import Path

from platformdirs import PlatformDirs

APP_NAME = "banteg/crimsonland"


def default_runtime_dir() -> Path:
    """Return the default per-user runtime directory.

    This is intended for saves/config/logs (e.g. `game.cfg`, `crimson.cfg`, highscores).
    Override with `CRIMSON_RUNTIME_DIR` (or legacy `CRIMSON_BASE_DIR`).
    """

    override = os.environ.get("CRIMSON_RUNTIME_DIR") or os.environ.get("CRIMSON_BASE_DIR")
    if override:
        return Path(override).expanduser()

    dirs = PlatformDirs(appname=APP_NAME, appauthor=False)
    return Path(dirs.user_data_path)



from __future__ import annotations

from grim.math import clamp

_RAD_TO_DEG = 57.29577951308232


def monster_vision_fade_alpha(hitbox_size: float) -> float:
    if float(hitbox_size) >= 0.0:
        return 1.0
    return clamp((float(hitbox_size) + 10.0) * 0.1, 0.0, 1.0)


__all__ = ["_RAD_TO_DEG", "monster_vision_fade_alpha"]



from __future__ import annotations

from collections.abc import Sequence

from ...sim.state_types import GameplayState, PlayerState
from ..ids import PerkId
from ..state import CreatureForPerks, PerkSelectionState
from .apply_context import PerkApplyCtx
from .counts import adjust_perk_count
from .manifest import PERK_APPLY_HANDLERS


def perk_apply(
    state: GameplayState,
    players: list[PlayerState],
    perk_id: PerkId,
    *,
    perk_state: PerkSelectionState | None = None,
    dt: float | None = None,
    creatures: Sequence[CreatureForPerks] | None = None,
) -> None:
    """Apply immediate perk effects and increment the perk counter."""

    if not players:
        return
    owner = players[0]
    try:
        adjust_perk_count(owner, perk_id)
        handler = PERK_APPLY_HANDLERS.get(perk_id)
        if handler is not None:
            handler(
                PerkApplyCtx(
                    state=state,
                    players=players,
                    owner=owner,
                    perk_id=perk_id,
                    perk_state=perk_state,
                    dt=dt,
                    creatures=creatures,
                ),
            )
    finally:
        if len(players) > 1:
            for player in players[1:]:
                player.perk_counts[:] = owner.perk_counts



from __future__ import annotations

from ..math_parity import f32
from .apply_context import BonusApplyCtx


def apply_weapon_power_up(ctx: BonusApplyCtx) -> None:
    old = float(ctx.state.bonuses.weapon_power_up)
    if old <= 0.0:
        ctx.register_global("weapon_power_up")
    ctx.state.bonuses.weapon_power_up = float(
        f32(float(old) + float(ctx.amount) * float(ctx.economist_multiplier)),
    )
    ctx.player.weapon_reset_latch = 0
    ctx.player.shot_cooldown = 0.0
    ctx.player.reload_active = False
    ctx.player.reload_timer = 0.0
    ctx.player.reload_timer_max = 0.0
    ctx.player.ammo = float(ctx.player.clip_size)



from __future__ import annotations

from dataclasses import dataclass

from grim.config import CrimsonConfig
from grim.geom import Vec2

from ...quests.types import format_level, parse_level

QUEST_MENU_BASE_X = -5.0
QUEST_MENU_BASE_Y = 185.0
QUEST_MENU_PANEL_OFFSET_X = -63.0

QUEST_TITLE_X_OFFSET = 219.0  # 300 + 64 - 145
QUEST_TITLE_Y_OFFSET = 44.0  # 40 + 4
QUEST_TITLE_W = 64.0
QUEST_TITLE_H = 32.0

QUEST_STAGE_ICON_X_OFFSET = 80.0  # 64 + 16
QUEST_STAGE_ICON_Y_OFFSET = 3.0
QUEST_STAGE_ICON_SIZE = 32.0
QUEST_STAGE_ICON_STEP = 36.0
QUEST_STAGE_ICON_SCALE_UNSELECTED = 0.8

QUEST_LIST_Y_OFFSET = 50.0
QUEST_LIST_ROW_STEP = 20.0
QUEST_LIST_NAME_X_OFFSET = 32.0
QUEST_LIST_HOVER_LEFT_PAD = 10.0
QUEST_LIST_HOVER_RIGHT_PAD = 210.0
QUEST_LIST_HOVER_TOP_PAD = 2.0
QUEST_LIST_HOVER_BOTTOM_PAD = 18.0

QUEST_HARDCORE_UNLOCK_INDEX = 40
QUEST_HARDCORE_CHECKBOX_X_OFFSET = 132.0
QUEST_HARDCORE_CHECKBOX_Y_OFFSET = -12.0
QUEST_HARDCORE_LIST_Y_SHIFT = 10.0

QUEST_BACK_BUTTON_X_OFFSET = 138.0
QUEST_BACK_BUTTON_Y_OFFSET = 212.0
QUEST_PANEL_HEIGHT = 378.0


@dataclass(frozen=True, slots=True)
class _QuestMenuLayout:
    title_pos: Vec2
    icons_start_pos: Vec2
    list_pos: Vec2


# game_update_victory_screen (0x00406350): used as the "end note" screen after the final quest.
END_NOTE_PANEL_POS_X = -45.0
END_NOTE_PANEL_POS_Y = 110.0
END_NOTE_PANEL_GEOM_X0 = -63.0
END_NOTE_PANEL_GEOM_Y0 = -81.0
END_NOTE_PANEL_W = 510.0
END_NOTE_PANEL_H = 378.0

END_NOTE_HEADER_X_OFFSET = 214.0  # v11 + 44 - 10 in the decompile, relative to panel-left
END_NOTE_HEADER_Y_OFFSET = 46.0  # (base_y + 40) + 6 in the decompile, relative to panel-top
END_NOTE_BODY_X_OFFSET = END_NOTE_HEADER_X_OFFSET - 8.0
END_NOTE_BODY_Y_GAP = 32.0
END_NOTE_LINE_STEP_Y = 14.0
END_NOTE_AFTER_BODY_Y_GAP = 22.0  # 14 + 8 in the decompile

END_NOTE_BUTTON_X_OFFSET = 266.0  # (v11 + 44 + 20) - 4 + 26, relative to panel-left
END_NOTE_BUTTON_Y_OFFSET = 210.0  # (base_y + 40) + 170 in the decompile, relative to panel-top
END_NOTE_BUTTON_STEP_Y = 32.0

# `quest_failed_screen_update` panel geometry/anchors:
# - panel is the classic ui_menuPanel at (-45, 110) with geom x0/y0 (-63, -81)
# - reaper banner X = panel-left + 214; message/buttons are derived from that anchor.
QUEST_FAILED_PANEL_POS_X = -45.0
QUEST_FAILED_PANEL_POS_Y = 110.0
QUEST_FAILED_PANEL_GEOM_X0 = -63.0
QUEST_FAILED_PANEL_GEOM_Y0 = -81.0
QUEST_FAILED_PANEL_W = 510.0
QUEST_FAILED_PANEL_H = 378.0

QUEST_FAILED_BANNER_X_OFFSET = 214.0
QUEST_FAILED_BANNER_Y_OFFSET = 40.0
QUEST_FAILED_BANNER_W = 256.0
QUEST_FAILED_BANNER_H = 64.0

QUEST_FAILED_MESSAGE_X_OFFSET = QUEST_FAILED_BANNER_X_OFFSET + 30.0
QUEST_FAILED_MESSAGE_Y_OFFSET = 126.0  # (base_y + 40) + 70 + 16
QUEST_FAILED_SCORE_X_OFFSET = QUEST_FAILED_BANNER_X_OFFSET + 40.0
QUEST_FAILED_SCORE_Y_OFFSET = 152.0  # message_y + 16 + 10 in native
QUEST_FAILED_BUTTON_X_OFFSET = QUEST_FAILED_BANNER_X_OFFSET + 52.0
QUEST_FAILED_BUTTON_Y_OFFSET = 240.0  # score_y baseline + 98 in native
QUEST_FAILED_BUTTON_STEP_Y = 32.0
QUEST_FAILED_PANEL_SLIDE_DURATION_MS = 250.0


def _player_name_default(config: CrimsonConfig) -> str:
    return config.player_name


def _next_quest_level(level: str) -> str | None:
    major, minor = parse_level(level)

    from ...quests import quest_by_stage

    for _ in range(100):
        minor += 1
        if minor > 10:
            minor = 1
            major += 1
        if quest_by_stage(major, minor) is not None:
            return format_level(major, minor)
    return None


__all__ = [
    "END_NOTE_AFTER_BODY_Y_GAP",
    "END_NOTE_BODY_X_OFFSET",
    "END_NOTE_BODY_Y_GAP",
    "END_NOTE_BUTTON_STEP_Y",
    "END_NOTE_BUTTON_X_OFFSET",
    "END_NOTE_BUTTON_Y_OFFSET",
    "END_NOTE_HEADER_X_OFFSET",
    "END_NOTE_HEADER_Y_OFFSET",
    "END_NOTE_LINE_STEP_Y",
    "END_NOTE_PANEL_GEOM_X0",
    "END_NOTE_PANEL_GEOM_Y0",
    "END_NOTE_PANEL_H",
    "END_NOTE_PANEL_POS_X",
    "END_NOTE_PANEL_POS_Y",
    "END_NOTE_PANEL_W",
    "QUEST_BACK_BUTTON_X_OFFSET",
    "QUEST_BACK_BUTTON_Y_OFFSET",
    "QUEST_FAILED_BANNER_H",
    "QUEST_FAILED_BANNER_W",
    "QUEST_FAILED_BANNER_X_OFFSET",
    "QUEST_FAILED_BANNER_Y_OFFSET",
    "QUEST_FAILED_BUTTON_STEP_Y",
    "QUEST_FAILED_BUTTON_X_OFFSET",
    "QUEST_FAILED_BUTTON_Y_OFFSET",
    "QUEST_FAILED_MESSAGE_X_OFFSET",
    "QUEST_FAILED_MESSAGE_Y_OFFSET",
    "QUEST_FAILED_PANEL_GEOM_X0",
    "QUEST_FAILED_PANEL_GEOM_Y0",
    "QUEST_FAILED_PANEL_H",
    "QUEST_FAILED_PANEL_POS_X",
    "QUEST_FAILED_PANEL_POS_Y",
    "QUEST_FAILED_PANEL_SLIDE_DURATION_MS",
    "QUEST_FAILED_PANEL_W",
    "QUEST_FAILED_SCORE_X_OFFSET",
    "QUEST_FAILED_SCORE_Y_OFFSET",
    "QUEST_HARDCORE_CHECKBOX_X_OFFSET",
    "QUEST_HARDCORE_CHECKBOX_Y_OFFSET",
    "QUEST_HARDCORE_LIST_Y_SHIFT",
    "QUEST_HARDCORE_UNLOCK_INDEX",
    "QUEST_LIST_HOVER_BOTTOM_PAD",
    "QUEST_LIST_HOVER_LEFT_PAD",
    "QUEST_LIST_HOVER_RIGHT_PAD",
    "QUEST_LIST_HOVER_TOP_PAD",
    "QUEST_LIST_NAME_X_OFFSET",
    "QUEST_LIST_ROW_STEP",
    "QUEST_LIST_Y_OFFSET",
    "QUEST_MENU_BASE_X",
    "QUEST_MENU_BASE_Y",
    "QUEST_MENU_PANEL_OFFSET_X",
    "QUEST_PANEL_HEIGHT",
    "QUEST_STAGE_ICON_SCALE_UNSELECTED",
    "QUEST_STAGE_ICON_SIZE",
    "QUEST_STAGE_ICON_STEP",
    "QUEST_STAGE_ICON_X_OFFSET",
    "QUEST_STAGE_ICON_Y_OFFSET",
    "QUEST_TITLE_H",
    "QUEST_TITLE_W",
    "QUEST_TITLE_X_OFFSET",
    "QUEST_TITLE_Y_OFFSET",
    "_QuestMenuLayout",
    "_next_quest_level",
    "_player_name_default",
]



from __future__ import annotations

from dataclasses import dataclass

from ..projectiles import ProjectileTypeId


@dataclass(frozen=True, slots=True)
class PlasmaProjectileRenderConfig:
    rgb: tuple[float, float, float]
    spacing: float
    seg_limit: int
    tail_size: float
    head_size: float
    head_alpha_mul: float
    aura_rgb: tuple[float, float, float]
    aura_size: float
    aura_alpha_mul: float


_DEFAULT_PLASMA_RENDER_CONFIG = PlasmaProjectileRenderConfig(
    rgb=(1.0, 1.0, 1.0),
    spacing=2.1,
    seg_limit=3,
    tail_size=12.0,
    head_size=16.0,
    head_alpha_mul=0.45,
    aura_rgb=(1.0, 1.0, 1.0),
    aura_size=120.0,
    aura_alpha_mul=0.15,
)


PLASMA_PROJECTILE_RENDER_CONFIG_BY_TYPE_ID: dict[int, PlasmaProjectileRenderConfig] = {
    int(ProjectileTypeId.PLASMA_RIFLE): PlasmaProjectileRenderConfig(
        rgb=(1.0, 1.0, 1.0),
        spacing=2.5,
        seg_limit=8,
        tail_size=22.0,
        head_size=56.0,
        head_alpha_mul=0.45,
        aura_rgb=(1.0, 1.0, 1.0),
        aura_size=256.0,
        aura_alpha_mul=0.3,
    ),
    int(ProjectileTypeId.PLASMA_MINIGUN): _DEFAULT_PLASMA_RENDER_CONFIG,
    int(ProjectileTypeId.PLASMA_CANNON): PlasmaProjectileRenderConfig(
        rgb=(1.0, 1.0, 1.0),
        spacing=2.6,
        seg_limit=18,
        tail_size=44.0,
        head_size=84.0,
        head_alpha_mul=0.45,
        aura_rgb=(1.0, 1.0, 1.0),
        aura_size=256.0,
        aura_alpha_mul=0.4,
    ),
    int(ProjectileTypeId.SPIDER_PLASMA): PlasmaProjectileRenderConfig(
        rgb=(0.3, 1.0, 0.3),
        spacing=_DEFAULT_PLASMA_RENDER_CONFIG.spacing,
        seg_limit=_DEFAULT_PLASMA_RENDER_CONFIG.seg_limit,
        tail_size=_DEFAULT_PLASMA_RENDER_CONFIG.tail_size,
        head_size=_DEFAULT_PLASMA_RENDER_CONFIG.head_size,
        head_alpha_mul=_DEFAULT_PLASMA_RENDER_CONFIG.head_alpha_mul,
        aura_rgb=(0.3, 1.0, 0.3),
        aura_size=_DEFAULT_PLASMA_RENDER_CONFIG.aura_size,
        aura_alpha_mul=_DEFAULT_PLASMA_RENDER_CONFIG.aura_alpha_mul,
    ),
    int(ProjectileTypeId.SHRINKIFIER): PlasmaProjectileRenderConfig(
        rgb=(0.3, 0.3, 1.0),
        spacing=_DEFAULT_PLASMA_RENDER_CONFIG.spacing,
        seg_limit=_DEFAULT_PLASMA_RENDER_CONFIG.seg_limit,
        tail_size=_DEFAULT_PLASMA_RENDER_CONFIG.tail_size,
        head_size=_DEFAULT_PLASMA_RENDER_CONFIG.head_size,
        head_alpha_mul=_DEFAULT_PLASMA_RENDER_CONFIG.head_alpha_mul,
        aura_rgb=(0.3, 0.3, 1.0),
        aura_size=_DEFAULT_PLASMA_RENDER_CONFIG.aura_size,
        aura_alpha_mul=_DEFAULT_PLASMA_RENDER_CONFIG.aura_alpha_mul,
    ),
}


def plasma_projectile_render_config(type_id: int) -> PlasmaProjectileRenderConfig:
    return PLASMA_PROJECTILE_RENDER_CONFIG_BY_TYPE_ID.get(int(type_id), _DEFAULT_PLASMA_RENDER_CONFIG)


BEAM_EFFECT_SCALE_BY_TYPE_ID: dict[int, float] = {
    int(ProjectileTypeId.ION_MINIGUN): 1.05,
    int(ProjectileTypeId.ION_RIFLE): 2.2,
    int(ProjectileTypeId.ION_CANNON): 3.5,
}


def beam_effect_scale(type_id: int) -> float:
    return float(BEAM_EFFECT_SCALE_BY_TYPE_ID.get(int(type_id), 0.8))


KNOWN_PROJ_RGB_BY_TYPE_ID: dict[int, tuple[int, int, int]] = {
    int(ProjectileTypeId.ION_RIFLE): (120, 200, 255),
    int(ProjectileTypeId.ION_MINIGUN): (120, 200, 255),
    int(ProjectileTypeId.ION_CANNON): (120, 200, 255),
    int(ProjectileTypeId.FIRE_BULLETS): (255, 170, 90),
    int(ProjectileTypeId.SHRINKIFIER): (160, 255, 170),
    int(ProjectileTypeId.BLADE_GUN): (240, 120, 255),
}


def known_proj_rgb(type_id: int) -> tuple[int, int, int]:
    return KNOWN_PROJ_RGB_BY_TYPE_ID.get(int(type_id), (240, 220, 160))




from __future__ import annotations

import hashlib
import tempfile
from pathlib import Path
from typing import SupportsInt, cast

from ..persistence.save_status import (
    QUEST_PLAY_COUNT,
    UNKNOWN_TAIL_SIZE,
    WEAPON_USAGE_COUNT,
    GameStatus,
    build_status_blob,
)
from .protocol import StatusSnapshot


def status_snapshot_from_status(status: GameStatus | None) -> StatusSnapshot:
    if status is None:
        return StatusSnapshot()

    def _mask_u32(value: object) -> int:
        try:
            return int(cast(SupportsInt, value)) & 0xFFFFFFFF
        except (TypeError, ValueError, OverflowError):
            return 0

    def _take_u32_list(raw: object, *, size: int) -> list[int]:
        if not isinstance(raw, list):
            return [0] * int(size)
        out: list[int] = []
        for value in raw[: int(size)]:
            out.append(_mask_u32(value))
        if len(out) < int(size):
            out.extend([0] * (int(size) - len(out)))
        return out

    data = getattr(status, "data", {}) or {}
    tail = data.get("unknown_tail", b"")
    if not isinstance(tail, (bytes, bytearray)):
        tail = b""
    tail = bytes(tail)
    if len(tail) != int(UNKNOWN_TAIL_SIZE):
        tail = tail[: int(UNKNOWN_TAIL_SIZE)] + (b"\x00" * max(0, int(UNKNOWN_TAIL_SIZE) - len(tail)))

    return StatusSnapshot(
        quest_unlock_index=int(data.get("quest_unlock_index", 0) or 0) & 0xFFFF,
        quest_unlock_index_full=int(data.get("quest_unlock_index_full", 0) or 0) & 0xFFFF,
        weapon_usage_counts=_take_u32_list(data.get("weapon_usage_counts"), size=int(WEAPON_USAGE_COUNT)),
        quest_play_counts=_take_u32_list(data.get("quest_play_counts"), size=int(QUEST_PLAY_COUNT)),
        mode_play_survival=_mask_u32(data.get("mode_play_survival", 0)),
        mode_play_rush=_mask_u32(data.get("mode_play_rush", 0)),
        mode_play_typo=_mask_u32(data.get("mode_play_typo", 0)),
        mode_play_other=_mask_u32(data.get("mode_play_other", 0)),
        game_sequence_id=_mask_u32(data.get("game_sequence_id", 0)),
        unknown_tail=tail,
    )


def hash_status_snapshot(snapshot: StatusSnapshot) -> str:
    """Stable hash for sanity-checking host/client status snapshots."""

    data = {
        "quest_unlock_index": int(snapshot.quest_unlock_index) & 0xFFFF,
        "quest_unlock_index_full": int(snapshot.quest_unlock_index_full) & 0xFFFF,
        "weapon_usage_counts": list(snapshot.weapon_usage_counts),
        "quest_play_counts": list(snapshot.quest_play_counts),
        "mode_play_survival": int(snapshot.mode_play_survival) & 0xFFFFFFFF,
        "mode_play_rush": int(snapshot.mode_play_rush) & 0xFFFFFFFF,
        "mode_play_typo": int(snapshot.mode_play_typo) & 0xFFFFFFFF,
        "mode_play_other": int(snapshot.mode_play_other) & 0xFFFFFFFF,
        "game_sequence_id": int(snapshot.game_sequence_id) & 0xFFFFFFFF,
        "unknown_tail": bytes(snapshot.unknown_tail),
    }

    weapon_counts = data["weapon_usage_counts"]
    if len(weapon_counts) != int(WEAPON_USAGE_COUNT):
        weapon_counts = weapon_counts[: int(WEAPON_USAGE_COUNT)] + [0] * max(0, int(WEAPON_USAGE_COUNT) - len(weapon_counts))
        data["weapon_usage_counts"] = weapon_counts[: int(WEAPON_USAGE_COUNT)]
    quest_counts = data["quest_play_counts"]
    if len(quest_counts) != int(QUEST_PLAY_COUNT):
        quest_counts = quest_counts[: int(QUEST_PLAY_COUNT)] + [0] * max(0, int(QUEST_PLAY_COUNT) - len(quest_counts))
        data["quest_play_counts"] = quest_counts[: int(QUEST_PLAY_COUNT)]

    tail = data["unknown_tail"]
    if len(tail) != int(UNKNOWN_TAIL_SIZE):
        tail = tail[: int(UNKNOWN_TAIL_SIZE)] + (b"\x00" * max(0, int(UNKNOWN_TAIL_SIZE) - len(tail)))
        data["unknown_tail"] = tail[: int(UNKNOWN_TAIL_SIZE)]

    digest = hashlib.sha256(build_status_blob(data)).hexdigest()
    return str(digest)


def build_lan_deterministic_status(*, snapshot: StatusSnapshot | None = None) -> GameStatus:
    """Return a session-local `GameStatus` for deterministic LAN simulation.

    The host sends its save snapshot in `MatchStart`, and all peers use it as the
    simulation status. This avoids split brain where local save progress impacts
    deterministic simulation (weapon availability / RNG consumption).
    """

    # This status object should never overwrite the on-disk save. Give it a
    # safe temp path so accidental `save_if_dirty()` calls don't clobber
    # `game.cfg`.
    path = Path(tempfile.gettempdir()) / "crimson-lan-sim-game.cfg"
    snap = snapshot or StatusSnapshot()
    weapon_counts = list(snap.weapon_usage_counts)
    if len(weapon_counts) != int(WEAPON_USAGE_COUNT):
        weapon_counts = weapon_counts[: int(WEAPON_USAGE_COUNT)] + [0] * max(0, int(WEAPON_USAGE_COUNT) - len(weapon_counts))
        weapon_counts = weapon_counts[: int(WEAPON_USAGE_COUNT)]
    quest_counts = list(snap.quest_play_counts)
    if len(quest_counts) != int(QUEST_PLAY_COUNT):
        quest_counts = quest_counts[: int(QUEST_PLAY_COUNT)] + [0] * max(0, int(QUEST_PLAY_COUNT) - len(quest_counts))
        quest_counts = quest_counts[: int(QUEST_PLAY_COUNT)]
    tail = bytes(snap.unknown_tail)
    if len(tail) != int(UNKNOWN_TAIL_SIZE):
        tail = tail[: int(UNKNOWN_TAIL_SIZE)] + (b"\x00" * max(0, int(UNKNOWN_TAIL_SIZE) - len(tail)))
        tail = tail[: int(UNKNOWN_TAIL_SIZE)]

    data = {
        "quest_unlock_index": int(snap.quest_unlock_index) & 0xFFFF,
        "quest_unlock_index_full": int(snap.quest_unlock_index_full) & 0xFFFF,
        "weapon_usage_counts": weapon_counts,
        "quest_play_counts": quest_counts,
        "mode_play_survival": int(snap.mode_play_survival) & 0xFFFFFFFF,
        "mode_play_rush": int(snap.mode_play_rush) & 0xFFFFFFFF,
        "mode_play_typo": int(snap.mode_play_typo) & 0xFFFFFFFF,
        "mode_play_other": int(snap.mode_play_other) & 0xFFFFFFFF,
        "game_sequence_id": int(snap.game_sequence_id) & 0xFFFFFFFF,
        "unknown_tail": tail,
    }
    return GameStatus(path=path, data=data, dirty=False)



from __future__ import annotations

from .end_note import EndNoteView
from .quest_failed import QUEST_FAILED_PANEL_SLIDE_DURATION_MS, QUEST_FAILED_PANEL_W, QuestFailedView
from .quest_results import QuestResultsView
from .quests_menu import QuestsMenuView

__all__ = [
    "EndNoteView",
    "QUEST_FAILED_PANEL_SLIDE_DURATION_MS",
    "QUEST_FAILED_PANEL_W",
    "QuestFailedView",
    "QuestResultsView",
    "QuestsMenuView",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from grim.fonts.small import SmallFontData

from ..frame import RenderFrame
from .bonuses import WorldRendererBonusesMixin
from .context import WorldRendererContextMixin
from .creatures import WorldRendererCreaturesMixin
from .draw import WorldRendererDrawMixin
from .effects import WorldRendererEffectsMixin
from .overlays import WorldRendererOverlaysMixin
from .projectiles import WorldRendererProjectilesMixin
from .trooper import WorldRendererTrooperMixin

if TYPE_CHECKING:
    from ...game_world import GameWorld


@dataclass(slots=True)
class WorldRenderer(
    WorldRendererDrawMixin,
    WorldRendererEffectsMixin,
    WorldRendererProjectilesMixin,
    WorldRendererTrooperMixin,
    WorldRendererCreaturesMixin,
    WorldRendererOverlaysMixin,
    WorldRendererBonusesMixin,
    WorldRendererContextMixin,
):
    _world: GameWorld
    _render_frame: RenderFrame | None = None
    _small_font: SmallFontData | None = None



from __future__ import annotations

import pyray as rl

from grim.geom import Vec2
from grim.math import clamp

from .constants import _RAD_TO_DEG
from .mixin_base import WorldRendererMixinBase


class WorldRendererOverlaysMixin(WorldRendererMixinBase):
    @staticmethod
    def _grim2d_circle_segments_filled(radius: float) -> int:
        # grim_draw_circle_filled (grim.dll): segments = trunc(radius * 0.125 + 12.0)
        return max(3, int(radius * 0.125 + 12.0))

    @staticmethod
    def _grim2d_circle_segments_outline(radius: float) -> int:
        # grim_draw_circle_outline (grim.dll): segments = trunc(radius * 0.2 + 14.0)
        return max(3, int(radius * 0.2 + 14.0))

    def _draw_aim_circle(self, *, center: Vec2, radius: float, alpha: float = 1.0) -> None:
        if radius <= 1e-3:
            return
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return

        fill_a = int(77 * alpha + 0.5)  # ui_render_aim_indicators: rgba(0,0,0.1,0.3)
        outline_a = int(255 * 0.55 * alpha + 0.5)
        fill = rl.Color(0, 0, 26, fill_a)
        outline = rl.Color(255, 255, 255, outline_a)

        rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)

        # The original uses a triangle fan (polygons). Raylib provides circle
        # primitives that still use triangles internally, but allow higher
        # segment counts for a smoother result when scaled.
        seg_count = max(self._grim2d_circle_segments_filled(radius), 64, int(radius))
        center_rl = center.to_rl()
        rl.draw_circle_sector(center_rl, float(radius), 0.0, 360.0, int(seg_count), fill)

        seg_count = max(self._grim2d_circle_segments_outline(radius), int(seg_count))
        # grim_draw_circle_outline draws a 2px-thick ring (outer radius = r + 2).
        # The exe binds bulletTrail, but that texture is white; the visual intent is
        # a subtle white outline around the filled spread circle.
        rl.draw_ring(center_rl, float(radius), float(radius + 2.0), 0.0, 360.0, int(seg_count), outline)

        rl.rl_set_texture(0)
        rl.end_blend_mode()

    def _draw_clock_gauge(self, *, pos: Vec2, ms: int, scale: float, alpha: float = 1.0) -> None:
        if self.clock_table_texture is None or self.clock_pointer_texture is None:
            return
        size = 32.0 * scale
        if size <= 1e-3:
            return
        tint = rl.Color(255, 255, 255, int(clamp(float(alpha), 0.0, 1.0) * 255.0 + 0.5))
        half = size * 0.5

        table_src = rl.Rectangle(
            0.0, 0.0, float(self.clock_table_texture.width), float(self.clock_table_texture.height),
        )
        table_dst = rl.Rectangle(pos.x, pos.y, size, size)
        rl.draw_texture_pro(self.clock_table_texture, table_src, table_dst, rl.Vector2(0.0, 0.0), 0.0, tint)

        seconds = int(ms) // 1000
        pointer_src = rl.Rectangle(
            0.0,
            0.0,
            float(self.clock_pointer_texture.width),
            float(self.clock_pointer_texture.height),
        )
        pointer_dst = rl.Rectangle(pos.x + half, pos.y + half, size, size)
        origin = rl.Vector2(half, half)
        rotation_deg = float(seconds) * 6.0
        rl.draw_texture_pro(self.clock_pointer_texture, pointer_src, pointer_dst, origin, rotation_deg, tint)

    def _hud_indicator_enabled(self, player_index: int) -> bool:
        if self.config is None:
            return True
        raw = self.config.hud_indicators
        if not isinstance(raw, (bytes, bytearray)):
            return True
        idx = int(player_index)
        if idx < 0:
            return False
        if idx >= len(raw):
            return True
        return bool(raw[idx])

    def _direction_arrow_tint(self, player_index: int, *, alpha: float) -> rl.Color:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if len(self.players) == 2:
            if int(player_index) == 0:
                return rl.Color(204, 230, 255, int(153.0 * alpha + 0.5))
            return rl.Color(255, 230, 204, int(153.0 * alpha + 0.5))
        return rl.Color(255, 255, 255, int(77.0 * alpha + 0.5))

    def _draw_direction_arrows(self, *, camera: Vec2, view_scale: Vec2, scale: float, alpha: float = 1.0) -> None:
        alpha = clamp(float(alpha), 0.0, 1.0)
        if alpha <= 1e-3:
            return
        arrow = self.arrow_texture
        if arrow is None:
            return

        src = rl.Rectangle(0.0, 0.0, float(arrow.width), float(arrow.height))
        width = max(1.0, float(arrow.width) * scale)
        height = max(1.0, float(arrow.height) * scale)
        origin = rl.Vector2(width * 0.5, height * 0.5)

        for player in self.players:
            if float(getattr(player, "health", 0.0)) <= 0.0:
                continue
            index = int(getattr(player, "index", 0))
            if not self._hud_indicator_enabled(index):
                continue

            heading = float(getattr(player, "heading", 0.0))
            marker_pos = player.pos + Vec2.from_heading(heading) * 60.0
            screen = self._world_to_screen_with(marker_pos, camera=camera, view_scale=view_scale)
            dst = rl.Rectangle(screen.x, screen.y, width, height)
            tint = self._direction_arrow_tint(index, alpha=alpha)
            rl.draw_texture_pro(arrow, src, dst, origin, float(heading * _RAD_TO_DEG), tint)



from __future__ import annotations

import random
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path

import pyray as rl

from . import music, sfx
from .config import CrimsonConfig
from .console import ConsoleState


@dataclass(slots=True)
class AudioState:
    ready: bool
    music: music.MusicState
    sfx: sfx.SfxState


def init_audio_state(config: CrimsonConfig, assets_dir: Path, console: ConsoleState) -> AudioState:
    music_disabled = config.music_disabled
    sound_disabled = config.sound_disabled
    music_volume = config.music_volume
    sfx_volume = config.sfx_volume

    music_enabled = not music_disabled
    sfx_enabled = not sound_disabled
    if not music_enabled and not sfx_enabled:
        console.log.log("audio: disabled (music + sfx)")
        console.log.flush()
        return AudioState(
            ready=False,
            music=music.init_music_state(ready=False, enabled=False, volume=music_volume),
            sfx=sfx.init_sfx_state(ready=False, enabled=False, volume=sfx_volume),
        )

    if not rl.is_audio_device_ready():
        rl.init_audio_device()
    ready = bool(rl.is_audio_device_ready())
    if not ready:
        console.log.log("audio: device init failed")
        console.log.flush()
        return AudioState(
            ready=False,
            music=music.init_music_state(ready=False, enabled=False, volume=music_volume),
            sfx=sfx.init_sfx_state(ready=False, enabled=False, volume=sfx_volume),
        )

    state = AudioState(
        ready=True,
        music=music.init_music_state(ready=True, enabled=music_enabled, volume=music_volume),
        sfx=sfx.init_sfx_state(ready=True, enabled=sfx_enabled, volume=sfx_volume),
    )
    sfx.load_sfx_index(state.sfx, assets_dir, console)
    music.load_music_tracks(state.music, assets_dir, console)
    return state


def play_music(state: AudioState, track_name: str) -> None:
    music.play_music(state.music, track_name)


def stop_music(state: AudioState) -> None:
    music.stop_music(state.music)


def trigger_game_tune(state: AudioState, *, rand: Callable[[], int] | None = None) -> str | None:
    return music.trigger_game_tune(state.music, rand=rand)


def play_sfx(
    state: AudioState | None,
    key: str | None,
    *,
    rng: random.Random | None = None,
    allow_variants: bool = True,
    reflex_boost_timer: float = 0.0,
) -> None:
    if state is None:
        return
    sfx.play_sfx(
        state.sfx,
        key,
        rng=rng,
        allow_variants=allow_variants,
        reflex_boost_timer=float(reflex_boost_timer),
    )


def set_sfx_volume(state: AudioState | None, volume: float) -> None:
    if state is None:
        return
    sfx.set_sfx_volume(state.sfx, volume)


def set_music_volume(state: AudioState | None, volume: float) -> None:
    if state is None:
        return
    music.set_music_volume(state.music, volume)


def update_audio(state: AudioState, dt: float) -> None:
    music.update_music(state.music, dt)


def shutdown_audio(state: AudioState) -> None:
    if not state.ready:
        return
    sfx.shutdown_sfx(state.sfx)
    music.shutdown_music(state.music)
    rl.close_audio_device()
